/*****************************************************************************
* if_wanipe_common.h   Sangoma Driver/Socket common area definitions.
*
* Author:       Nenad Corbic <ncorbic@sangoma.com>
*
* Copyright:    (c) 2000 Sangoma Technologies Inc.
*
*               This program is free software; you can redistribute it and/or
*               modify it under the terms of the GNU General Public License
*               as published by the Free Software Foundation; either version
*               2 of the License, or (at your option) any later version.
* ============================================================================
* Jan 13, 2000  Nenad Corbic      Initial version
*****************************************************************************/


#ifndef _WANPIPE_SOCK_DRIVER_COMMON_H
#define _WANPIPE_SOCK_DRIVER_COMMON_H

#if defined(WAN_KERNEL)

#if defined(__FreeBSD__) || defined(__OpenBSD__)
# include <net/wanpipe_debug.h>
# include <net/wanpipe_common.h>
# include <net/wanpipe_kernel.h>
#elif defined(__LINUX__)
# include <linux/version.h>
# include <linux/wanpipe_debug.h>
# include <linux/wanpipe_common.h>
# include <linux/wanpipe_kernel.h>
#endif

#define wan_next_dev(dev)	((wanpipe_common_t*)wan_netif_priv(dev))->next
#define wan_prot_dev(dev)	((wanpipe_common_t*)wan_netif_priv(dev))->prot_ptr

#if defined(__FreeBSD__) || defined(__OpenBSD__)
typedef struct {
	void*		prot_ptr;
	netdevice_t*	next;		/* netdevice_t 	*slave; */
	sdla_t*		card;
	netdevice_t*	dev;
	unsigned char	state;
	unsigned char	usedby;
#if defined(CONFIG_PRODUCT_WANPIPE_GENERIC)
	unsigned int	proto;
#endif
} wanpipe_common_t;
#elif defined(__LINUX__)
typedef struct {
	void*		prot_ptr;
	netdevice_t*	next;		/* netdevice_t 	*slave; */
	void*		card;		/* ALEX */
	
	atomic_t receive_block;
	atomic_t command;
	atomic_t disconnect;
	
	struct sock *sk;		/* Wanpipe Sock bind's here */ 
	
	struct tq_struct wanpipe_task;	/* Immediate BH handler task */
	
	unsigned char rw_bind;		/* Sock bind state */
	unsigned char usedby;
	unsigned char state;
	unsigned char svc;
	unsigned short lcn;
	unsigned int config_id;
	
	unsigned long used;
	unsigned long api_state;
	netdevice_t *dev;
	struct sk_buff_head rx_queue;
	wan_tasklet_t  task;
	struct timer_list dev_timer;
#if defined(CONFIG_PRODUCT_WANPIPE_GENERIC)
	unsigned int	proto;
#endif
} wanpipe_common_t;

/* Used flags: Resources */
enum {
LCN_USED,
LCN_DEV,
LCN_TX_DEV,
LCN_X25_LINK,
LCN_SK_ID,
LCN_DSP_ID,
WAN_API_INIT
};

#define SK_ID LCN_SK_ID


enum {
	WANSOCK_UNCONFIGURED,	/* link/channel is not configured */
	WANSOCK_DISCONNECTED,	/* link/channel is disconnected */
	WANSOCK_CONNECTING,		/* connection is in progress */
	WANSOCK_CONNECTED,		/* link/channel is operational */
	WANSOCK_LIMIT,		/* for verification only */
	WANSOCK_DUALPORT,		/* for Dual Port cards */
	WANSOCK_DISCONNECTING,
	WANSOCK_BINDED,
	WANSOCK_BIND_LISTEN,
	WANSOCK_LISTEN
};

#endif /* __LINUX__ */

#endif /* WAN_KERNEL */

#endif


