/*
 * Copyright (c) 2002
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 *	$Id: wanpipe_common.h,v 1.70 2004/02/23 13:45:41 sangoma Exp $
 */

/****************************************************************************
 * wanpipe_common.h	WANPIPE(tm) Multiprotocol WAN Link Driver.
 *
 * Author: 	Alex Feldman <al.feldman@sangoma.com>
 *
 * ==========================================================================
 * July 17, 2002 Alex Feldman 	Initial Version
 ****************************************************************************
 */

#ifndef	__WANPIPE_COMMON_H
# define	__WANPIPE_COMMON_H

#ifdef __LINUX__
# include <linux/wanpipe_kernel.h>
#endif

#ifdef WAN_DEBUG_MEM
extern atomic_t wan_debug_mem;
#endif

/*
*****************************************************************************
**			D E F I N E S				
*****************************************************************************
*/
#ifndef NIPQUAD
# define NIPQUAD(addr) \
	((unsigned char *)&addr)[0], \
	((unsigned char *)&addr)[1], \
	((unsigned char *)&addr)[2], \
	((unsigned char *)&addr)[3]
#endif

#if defined(__FreeBSD__)
# define WAN_TAILQ_FIRST(ifp)	TAILQ_FIRST(&ifp->if_addrhead)
# define WAN_TAILQ_NEXT(ifa)	TAILQ_NEXT(ifa, ifa_link)
#elif defined (__OpenBSD__)
# define WAN_TAILQ_FIRST(ifp)	TAILQ_FIRST(&ifp->if_addrlist)
# define WAN_TAILQ_NEXT(ifa)	TAILQ_NEXT(ifa, ifa_list)
#elif defined(__LINUX__)
#elif defined(__WINDOWS__)
#else
# error "WAN_TAILQ_x macros doesn't supported yet!"
#endif

#if defined(__FreeBSD__) || defined(__OpenBSD__)
# ifndef IFQ_SET_READY
#  define IFQ_SET_READY(ifq)		/* nothing */
# endif
# ifndef IFQ_IS_EMPTY
#  define IFQ_IS_EMPTY(ifq)		((ifq)->ifq_len == 0)
# endif
# ifndef IFQ_INC_LEN
#  define IFQ_INC_LEN(ifq)		((ifq)->ifq_len++)
# endif
# ifndef IFQ_DEC_LEN
#  define IFQ_DEC_LEN(ifq)		(--(ifq)->ifq_len)
# endif
# ifndef IFQ_INC_DROPS
#  define IFQ_INC_DROPS(ifq)		((ifq)->ifq_drops++)
# endif
# ifndef IFQ_SET_MAXLEN
#  define IFQ_SET_MAXLEN(ifq, len)	((ifq)->ifq_maxlen = (len))
# endif
# if (__FreeBSD_version > 500000)
#  define IF_DROP	_IF_DROP
#  define IF_QFULL	_IF_QFULL
# endif
# ifndef IFQ_PURGE
#  if (__FreeBSD_version > 500000)
#   define IFQ_PURGE(ifq)	IF_DRAIN((ifq))
#  else
#   define IFQ_PURGE(ifq)					\
	while (1) {						\
		struct mbuf *m0;				\
		IF_DEQUEUE((ifq), m0);				\
		if (m0 == NULL)					\
			break;					\
		else						\
			m_freem(m0);				\
	}
#  endif
# endif
# ifndef IFQ_ENQUEUE
#  if (__FreeBSD_version > 500000)
#   define IFQ_ENQUEUE(ifq, m, pattr, err)	IF_ENQUEUE((ifq), (m))
#  else
#   define IFQ_ENQUEUE(ifq, m, pattr, err)			\
	do {							\
		if (IF_QFULL((ifq))) {				\
			m_freem((m));				\
			(err) = ENOBUFS;			\
		} else {					\
			IF_ENQUEUE((ifq), (m));			\
			(err) = 0;				\
		}						\
		if ((err))					\
			(ifq)->ifq_drops++;			\
	} while (0)
#  endif
# endif
# ifndef IFQ_DEQUEUE
#  define IFQ_DEQUEUE(ifq, m)	IF_DEQUEUE((ifq), (m))
# endif
# ifndef IFQ_POLL
#  define IFQ_POLL(ifq, m)	((m) = (ifq)->ifq_head)
# endif
# ifndef IFQ_CLASSIFY
struct altq_pktattr { void* attr; };
#  define IFQ_CLASSIFY(ifq, m, sa_family, pktattr) (pktattr)->attr = NULL
# endif
# if (__FreeBSD_version > 500000)
#  define IFQ_INIT(ifq, max_pkt)		\
		IFQ_SET_MAXLEN((ifq), max_pkt);	\
		mtx_init(&(ifq)->ifq_mtx, "ifq", "wan", MTX_DEF);	\
 		(ifq)->ifq_len = 0;
# else
#  define IFQ_INIT(ifq, max_pkt)		\
		IFQ_SET_MAXLEN((ifq), max_pkt);	\
 		(ifq)->ifq_len = 0;
# endif
# define IFQ_LEN(ifq)	((ifq)->ifq_len)
#elif defined(__LINUX__)
# define IFQ_INIT(trace_queue, max_pkt)			skb_queue_head_init(trace_queue)
# define IFQ_PURGE(trace_queue)				skb_queue_purge(trace_queue)
# define IFQ_ENQUEUE(trace_queue, skb, arg, err)	skb_queue_tail(trace_queue, skb)
# define IFQ_LEN(trace_queue)				skb_queue_len(trace_queue)
#elif defined(__WINDOWS__)
#else
# error "Undefined IFQ_x macros!"
#endif

#if defined(__FreeBSD__)
# if (__FreeBSD_version < 410000)
#  define WAN_TASKLET_INIT(task, priority, func, arg)	\
	task->running = 0;				\
	task->task_func = func; task->data = arg
#  define WAN_TASKLET_SCHEDULE(task)					\
	if (!test_bit(0, &task->running)){				\
		set_bit(0, &task->running);				\
		task->task_func(task->data, 0);				\
	}
#  define WAN_TASKLET_END(task)	clear_bit(0, &task->running)
#  define WAN_TASKLET_KILL(task)
# else
#  define WAN_TASKLET_INIT(task, priority, func, arg)	\
	task->running = 0;				\
	TASK_INIT(&task->task_id, priority, func, arg)
#  define WAN_TASKLET_SCHEDULE(task)					\
	if (!test_bit(0, &task->running)){				\
		set_bit(0, &task->running);				\
		taskqueue_enqueue(taskqueue_swi, &task->task_id);	\
	}
/*		taskqueue_run(taskqueue_swi);				\*/
#  define WAN_TASKLET_END(task)	clear_bit(0, &task->running)
#  define WAN_TASKLET_KILL(task)
# endif
#elif defined(__OpenBSD__)
# define WAN_TASKLET_INIT(task, priority, func, arg)	\
	task->running = 0;				\
	task->task_func = func; task->data = arg
# define WAN_TASKLET_SCHEDULE(task)					\
	if (!test_bit(0, &task->running)){				\
		set_bit(0, &task->running);				\
		task->task_func(task->data, 0);				\
	}
# define WAN_TASKLET_END(task)	clear_bit(0, &task->running)
# define WAN_TASKLET_KILL(task)
#elif defined(__LINUX__)
# define WAN_TASKLET_INIT(task, priority, func, arg)	\
	task->running = 0;				\
	tasklet_init(&task->task_id,func,(unsigned long)arg) 
# define WAN_TASKLET_SCHEDULE(task)					\
	tasklet_hi_schedule(&task->task_id);			
# define WAN_TASKLET_END(task)	clear_bit(0, &task->running)
# define WAN_TASKLET_KILL(task)	tasklet_kill(&task->task_id)
#elif defined(__WINDOWS__)
#else
# error "Undefined WAN_TASKLET_x macro!"
#endif

#if defined(__FreeBSD__)
# if (__FreeBSD_version < 410000)
#  define WAN_TASKQ_INIT(task, priority, func, arg)	\
		task->tfunc = func; task->data = arg
# else
#  define WAN_TASKQ_INIT(task, priority, func, arg)	\
		TASK_INIT(&task->tqueue, priority, func, arg)
# endif
#elif defined(__OpenBSD__)
# define WAN_TASKQ_INIT(task, priority, func, arg)	\
		task->tfunc = func; task->data = arg
#elif defined(__LINUX__)
# define WAN_TASKQ_INIT(task, priority, func, arg)	\
		INIT_WORK((&task->tqueue),func,arg)
#elif defined(__WINDOWS__)
#else
# error "Undefined WAN_TASKQ_INIT macro!"
#endif

#if defined(__FreeBSD__) && (__FreeBSD_version >= 410000)
# define WAN_TASKQ_SCHEDULE(task)			\
	taskqueue_enqueue(taskqueue_swi, &task->tqueue);\
	taskqueue_run(taskqueue_swi)
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
# define WAN_TASKQ_SCHEDULE(task)			\
	task->tfunc(task->data, 0)
#elif defined(__LINUX__)
# define WAN_TASKQ_SCHEDULE(task)			\
	wan_schedule_task(&task->tqueue)
#elif defined(__WINDOWS__)
#else
# error "Undefined WAN_TASKQ_SCHEDULE macro!"
#endif

#if defined(__LINUX__)
# define WAN_COPY_FROM_USER(k,u,l)	copy_from_user(k,u,l)
# define WAN_COPY_TO_USER(u,k,l)	copy_to_user(u,k,l)
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
# define WAN_COPY_FROM_USER(k,u,l)	copyin(u,k,l)
# define WAN_COPY_TO_USER(u,k,l)	copyout(k,u,l)
#elif defined(__WINDOWS__)
#else
# error "Undefined WAN_COPY_FROM_USER/WAN_COPY_TO_USER macros!"
#endif

#if defined(__LINUX__)
# if (LINUX_VERSION_CODE < KERNEL_VERSION(2,3,43))
#  define WAN_NETIF_WAKE_QUEUE(dev)	do {				\
					clear_bit(0, &dev->tbusy);	\
					mark_bh(NET_BH);		\
					} while(0)
#  define WAN_NETIF_START_QUEUE(dev)	do {				\
					dev->tbusy = 0;			\
					dev->interrupt = 0;		\
					dev->start = 1;			\
					} while(0);
#  define WAN_NETIF_STOP_QUEUE(dev)	set_bit(0, &dev->tbusy)
#  define WAN_NETIF_RUNNING(dev)	dev->start
#  define WAN_NETDEVICE_START(dev)	dev->start = 1
#  define WAN_NETDEVICE_STOP(dev)	dev->start = 0
#  define WAN_NETIF_QUEUE_STOPPED(dev)	test_bit(0,&dev->tbusy)
# else
#  define WAN_NETIF_WAKE_QUEUE(dev)	netif_wake_queue(dev)
#  define WAN_NETIF_START_QUEUE(dev)	netif_start_queue(dev)
#  define WAN_NETIF_STOP_QUEUE(dev)	netif_stop_queue(dev)
#  define WAN_NETIF_RUNNING(dev)	netif_running(dev)
#  define WAN_NETDEVICE_START(dev)
#  define WAN_NETDEVICE_STOP(dev)
#  define WAN_NETIF_QUEUE_STOPPED(dev)	netif_queue_stopped(dev)
# endif
# define WAN_NETIF_UP(dev)		((dev)->flags&IFF_UP)
# define WAN_NET_RATELIMIT		net_ratelimit

#elif defined(__FreeBSD__) || defined(__OpenBSD__)
# define WAN_NETIF_QUEUE_STOPPED(dev)	0
# define WAN_NETIF_WAKE_QUEUE(dev)
# define WAN_NETIF_STOP_QUEUE(dev)	
# define WAN_NETDEVICE_STOP(dev)
# define WAN_NETDEVICE_START(dev)
# define WAN_NETIF_RUNNING(dev)	
# define WAN_NETIF_START_QUEUE(dev)
# define MOD_INC_USE_COUNT
# define MOD_DEC_USE_COUNT
# define WAN_NETIF_UP(dev)		((dev)->if_flags&IFF_UP)
# define NET_ADMIN_CHECK()
# define WAN_NET_RATELIMIT()		1
#elif defined(__WINDOWS__)
#else
# error "Undefined WAN_NETIF_x macros!"
#endif

#if defined(__LINUX__)
# define WAN_SPIN_LOCK_INIT(SpinLock)		\
		spin_lock_init(&((SpinLock)->slock))
# define WAN_SPIN_LOCK_IRQSAVE(SpinLock)	\
		spin_lock_irqsave(&((SpinLock)->slock), (SpinLock)->flags)
# define WAN_SPIN_UNLOCK_IRQSAVE(SpinLock)	\
		spin_unlock_irqrestore(&((SpinLock)->slock), (SpinLock)->flags)

# define WAN_RWLOCK_INIT(rwlock_ptr) \
		rwlock_ptr->rwlock = RW_LOCK_UNLOCKED;
	
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
# define WAN_SPIN_LOCK_INIT(SpinLock)	(SpinLock)->slock = 0
# define WAN_SPIN_LOCK_IRQSAVE(SpinLock)	\
		(SpinLock)->slock = splimp();
# define WAN_SPIN_UNLOCK_IRQSAVE(SpinLock)	\
		splx((SpinLock)->slock);
#elif defined(__WINDOWS__)
#else
# error "Undefined WAN_SPIN_LOCK_x macros!"
#endif

#if defined(__LINUX__)
# define WAN_BPF_REPORT(dev,m)
#elif defined(__FreeBSD__)
# if (__FreeBSD_version > 500000)
#  define WAN_BPF_REPORT(dev,m)	bpf_mtap((dev)->if_bpf, (m))
# else
#  define WAN_BPF_REPORT(dev,m)	bpf_mtap((dev), (m))
# endif
#elif defined(__OpenBSD__)
#  define WAN_BPF_REPORT(dev,m)	bpf_mtap((dev)->if_bpf, (m));
#elif defined(__WINDOWS__)
#else
# error "Undefined WAN_BPF_REPORT macro!"
#endif


#if defined (__LINUX__)
# define __WAN_PUT(str)	atomic_dec(&(str)->refcnt)
# define WAN_PUT(str)	if (atomic_dec_and_test(&(str)->refcnt)){ \
		       		wan_kfree(str); 		\
                       	} 
# define WAN_HOLD(str)	atomic_inc(&(str)->refcnt)
#elif defined(__FreeBSD__)
# define __WAN_PUT(str)	atomic_dec(&(str)->refcnt)
# define WAN_PUT(str)	atomic_dec(&str->refcnt);	\
			if (str->refcnt){		\
		       		WAN_FREE(str);		\
                       	} 
# define WAN_HOLD(str)	atomic_inc(&(str)->refcnt)
#elif defined(__OpenBSD__)
# define __WAN_PUT(str)	str->refcnt--
# define WAN_PUT(str)	str->refcnt--;	\
			if (str->refcnt){		\
		       		WAN_FREE(str);		\
                       	} 
# define WAN_HOLD(str)	str->refcnt++
#elif defined(__WINDOWS__)
#else
# warning "Undefined WAN_HOLD/WAN_PUT macro!"
#endif

#if defined(__FreeBSD__) || defined(__OpenBSD__)
# ifdef ENABLE_SPPP
#  define WAN_SPPP_ENABLED		1
#  define WAN_SPPP_ATTACH(ifp)		sppp_attach(ifp)
#  define WAN_SPPP_DETACH(ifp)		sppp_detach(ifp)
#  define WAN_SPPP_FLUSH(ifp)		sppp_flush(ifp)
#  define WAN_SPPP_PICK(ifp)		sppp_pick(ifp)
#  define WAN_SPPP_DEQUEUE(ifp)		sppp_dequeue(ifp)
#  define WAN_SPPP_ISEMPTY(ifp)		sppp_isempty(ifp)
#  define WAN_SPPP_INPUT(ifp,skb)	sppp_input(ifp,skb)
#  define WAN_SPPP_IOCTL(ifp,cmd,data)	sppp_ioctl(ifp,cmd,data);
# else
#  define WAN_SPPP_ENABLED		0
#  define WAN_SPPP_ATTACH(ifp)		
#  define WAN_SPPP_DETACH(ifp)		
#  define WAN_SPPP_FLUSH(ifp)		
#  define WAN_SPPP_PICK(ifp)		NULL
#  define WAN_SPPP_DEQUEUE(ifp)		NULL
#  define WAN_SPPP_ISEMPTY(ifp)		0
#  define WAN_SPPP_INPUT(ifp,skb)	
#  define WAN_SPPP_IOCTL(ifp,cmd,data)	-EOPNOTSUPP
# endif
#elif defined(__LINUX__)
# define WAN_SPPP_ENABLED		1
# define WAN_SPPP_IOCTL(ifp,cmd,data)	-EOPNOTSUPP
#elif defined(__WINDOWS__)
#else
# error "Undefined WAN_SPPP_x macros!"
#endif

#if defined(__FreeBSD__) || defined(__OpenBSD__)
# ifdef NBPFILTER
#  if defined(__FreeBSD__)
#   define BPF_ATTACH(ifp,type)	bpfattach(ifp, type, 4)
#  elif defined(__OpenBSD__)
#   define BPF_ATTACH(ifp,type)	bpfattach(&ifp->if_bpf, ifp, DLT_NULL, 4)
#  else
#   error "BPF_ATTACH macro doesn't supported yet!"
#  endif
# else
#  define BPF_ATTACH(ifp, type)
# endif
#endif

/*
*****************************************************************************
**			T Y P E D E F S				
*****************************************************************************
*/

/*
*****************************************************************************
**		F U N C T I O N   P R O T O T Y P E S				
*****************************************************************************
*/
char*		wanpipe_get_state_string (void*);
void 		wanpipe_set_state (void*, int);
char 		wanpipe_get_state (void*);
void		wanpipe_card_lock_irq (void *,unsigned long *);
void		wanpipe_card_unlock_irq (void *,unsigned long *);
void		wanpipe_set_baud(void*card,unsigned int baud);
unsigned long 	wan_get_ip_addr   (void*, int);
int		wan_udp_pkt_type  (void*, caddr_t);
int		wan_reply_udp	  (void*, unsigned char*, unsigned int);
void 		wanpipe_debug_timer_init(void*);
#if defined(__FreeBSD__) || defined(__OpenBSD__)
void		wanpipe_debugging (void* data, int pending);
#else
void		wanpipe_debugging (unsigned long data);
#endif

/*
*****************************************************************************
**		I N L I N E   F U N C T I O N S				
*****************************************************************************
*/
static __inline void wan_netif_free(netdevice_t*);
static __inline int wan_netif_init(netdevice_t*, char*, int);
static __inline int wan_netif_del(netdevice_t* dev);
/******************* WANPIPE MALLOC/FREE FUNCTION ******************/
/*
** wan_malloc - 
*/
static __inline void* wan_malloc(int size)
{
	void*	ptr = NULL;
#if defined(__LINUX__)
	ptr = kmalloc(size, GFP_ATOMIC);
	if (ptr){
		DEBUG_ADD_MEM(size);
	}
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	ptr = malloc(size, M_DEVBUF, M_NOWAIT); 
#elif defined(__WINDOWS__)
	ptr = ExAllocatePool(NonPagedPool, size);
#else
# error "wan_malloc() function is not supported yet!"
#endif
	if (ptr){
		memset(ptr, 0, size);
	}
	return ptr;
}

/*
** wan_free - 
*/
static __inline void wan_free(void* ptr)
{
	if (!ptr){
		DEBUG_EVENT("wan_free: NULL PTR !!!!!\n");
		return;
	}
	
#if defined(__LINUX__)
	kfree(ptr);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	return free(ptr, M_DEVBUF); 
#elif defined(__WINDOWS__)
	ExFreePool(ptr);
#else
# error "wan_free() function is not supported yet!"
#endif
}



/******************* WANPIPE VIRT<->BUS SPACE FUNCTION ******************/
/*
** wan_virt2bus
*/
static __inline unsigned long wan_virt2bus(unsigned long* ptr)
{
#if defined(__LINUX__)
	return virt_to_bus(ptr);
#elif defined(__FreeBSD__)
	return vtophys((vm_offset_t)ptr);
#elif defined(__OpenBSD__)
	return vtophys((vaddr_t)ptr);
#elif defined(__WINDOWS__)
#else
# error "wan_virt2bus() function is not supported yet!"
#endif
}

/*
** wan_bus2virt
*/
static __inline unsigned long* wan_bus2virt(unsigned long virt_addr)
{
#if defined(__LINUX__)
	return bus_to_virt(virt_addr);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	return (unsigned long*)virt_addr;
#elif defined(__WINDOWS__)
#else
# error "wan_bus2virt() function is not supported yet!"
#endif
}


/******************* WANPIPE DMA FUNCTION ******************/

/*
** wan_dma_alloc
*/
static __inline int
wan_dma_alloc(void* hw, wan_dma_descr_t* dma_descr)
{
	int	err = 0;
#if defined(__FreeBSD__)
	err = bus_dma_tag_create(/*parent*/NULL, 
				  /*alignemnt*/1, 
				  /*boundary*/0,
			          /*lowaddr*/BUS_SPACE_MAXADDR_32BIT,
			          /*highaddr*/BUS_SPACE_MAXADDR,
			          /*filter*/NULL, /*filterarg*/NULL,
			          /*maxsize*/dma_descr->max_length,
			          /*nsegments*/BUS_SPACE_UNRESTRICTED,
			          /*maxsegsz*/BUS_SPACE_MAXSIZE_32BIT,
			          /*flags*/0, 
				  (bus_dma_tag_t*)&dma_descr->dmat);
	if (err){
		DEBUG_EVENT("Failed create DMA tag (size=%ld)!\n",
					dma_descr->max_length);
		dma_descr->max_length = 0;
		return err;
	}
	err = bus_dmamem_alloc(dma_descr->dmat, 
			       (void**)&dma_descr->vAddr, 
			       BUS_DMA_NOWAIT, 
			       &dma_descr->dmamap);
	if (err){
		DEBUG_EVENT("Failed allocate DMA (size=%ld)!\n",
					dma_descr->max_length);
		bus_dma_tag_destroy(dma_descr->dmat);
		dma_descr->max_length = 0;
		return err;
	}
#elif defined(__OpenBSD__)
	err = bus_dmamem_alloc(dma_descr->dmat, 	/* dma tag */
				dma_descr->max_length, 	/* size */
				PAGE_SIZE, 		/* alignment */
				0,			/* boundary */
				&dma_descr->dmaseg,	/* serments */
				1,			/* num of segments */
				&dma_descr->rsegs,	/* R num of segments */
				BUS_DMA_NOWAIT);	
	if (err){
		DEBUG_EVENT("Failed allocate DMA segment (size=%ld)!\n",
					dma_descr->max_length);
		dma_descr->max_length = 0;
		return err;
	}
	err = bus_dmamem_map(dma_descr->dmat,		/* dma tag */
				&dma_descr->dmaseg,	/* segments */
				dma_descr->rsegs,	/* return num of segments */
				dma_descr->max_length, 	/* size */
				(caddr_t*)&dma_descr->vAddr,	/* kernel virtual address */
				BUS_DMA_NOWAIT);
	if (err){
		DEBUG_EVENT("Failed map DMA segment (size=%ld)!\n",
					dma_descr->max_length);
		dma_descr->max_length = 0;
		bus_dmamem_free(dma_descr->dmat, &dma_descr->dmaseg, dma_descr->rsegs);
		return err;
	}		
#elif defined(__LINUX__)
	dma_descr->vAddr = pci_alloc_consistent(NULL,dma_descr->max_length,(dma_addr_t *)&dma_descr->pAddr); 
	if (dma_descr->vAddr == NULL){
		err = -ENOMEM;
	}
#elif defined(__WINDOWS__)
	return -EINVAL;
#else
# error "wan_dma_alloc() function is not supported yet!"
#endif
	return err;
}

/*
** wan_dma_free
*/
static __inline int
wan_dma_free(void* hw, wan_dma_descr_t* dma_descr)
{
#if defined(__FreeBSD__)
	bus_dmamem_free(dma_descr->dmat, dma_descr->vAddr, dma_descr->dmamap);
	return bus_dma_tag_destroy(dma_descr->dmat);
#elif defined(__OpenBSD__)
	bus_dmamem_unmap(dma_descr->dmat, (caddr_t)dma_descr->vAddr, dma_descr->max_length);
	bus_dmamem_free(dma_descr->dmat, &dma_descr->dmaseg, dma_descr->rsegs);
#elif defined(__LINUX__)

	DEBUG_TEST("Freeing Pages 0x%p len=%li  order=%i\n",
		dma_descr->vAddr,
		dma_descr->max_length,
		get_order(dma_descr->max_length));

	pci_free_consistent(NULL, dma_descr->max_length,dma_descr->vAddr,dma_descr->pAddr);
	dma_descr->vAddr = NULL;
	dma_descr->pAddr = 0;
#elif defined(__WINDOWS__)
	return -EINVAL;
#else
# error "wan_dma_free() function is not supported yet!"
#endif
	return 0;
}


/********************** WANPIPE TIMER FUNCTION **************************/


static __inline unsigned long wan_getcurrenttime(void)
{
#if defined(__FreeBSD__) || defined(__OpenBSD__)
	struct timeval 	tv;
	microtime(&tv);
	return tv.tv_sec;
#elif defined(__WINDOWS__)
	LARGE_INTEGER	tv;
	NdisGetCurrentSystemTime(&tv);
	return (unsigned long)tv.QuadPart;
#elif defined(__LINUX__)
	struct timeval 	tv;
	do_gettimeofday(&tv);
	return tv.tv_sec;
#else
# error "wan_getcurrenttime() function is not supported yet!"
#endif
}

/*
** wan_init_timer
*/
static __inline void
wan_init_timer(wan_timer_t* wan_timer, wan_timer_func_t timer_func, unsigned long arg)
{
#if defined(__LINUX__)
	init_timer(&wan_timer->timer_info);
	wan_timer->timer_info.function = timer_func;
	wan_timer->timer_info.data = arg;
#elif defined(__FreeBSD__)
	/* FIXME_ADSL_TIMER */
	callout_handle_init(&wan_timer->timer_info);
	wan_timer->timer_func = timer_func;
	wan_timer->timer_arg  = (void*)arg;
#elif defined(__OpenBSD__)
	timeout_set(&wan_timer->timer_info, timer_func, (void*)arg);
	wan_timer->timer_func = timer_func;
	wan_timer->timer_arg  = (void*)arg;
#elif defined(__WINDOWS__)
#else
# error "wan_init_timer() function is not supported yet!"
#endif /* linux */
}

/*
** wan_del_timer
*/
static __inline void
wan_del_timer(wan_timer_t* wan_timer)
{
#if defined(__LINUX__)
	del_timer(&wan_timer->timer_info);
#elif defined(__FreeBSD__)
	untimeout(wan_timer->timer_func,
		  (void*)wan_timer->timer_arg, 
		  wan_timer->timer_info);
	callout_handle_init(&wan_timer->timer_info);
#elif defined(__OpenBSD__)
	timeout_del(&wan_timer->timer_info);
#else
# error "wan_del_timer() function is not supported yet!"
#endif /* linux */
}

/*
** wan_add_timer
*/
static __inline void 
wan_add_timer(wan_timer_t* wan_timer, unsigned long delay)
{
#if defined(__LINUX__)
    	wan_timer->timer_info.expires = SYSTEM_TICKS + delay;
    	add_timer(&wan_timer->timer_info);
#elif defined(__FreeBSD__)
	wan_timer->timer_info = 
			timeout(wan_timer->timer_func,
				(void*)wan_timer->timer_arg, 
				delay);
	WAN_ASSERT1(wan_timer->timer_info.callout == NULL);
#elif defined(__OpenBSD__)
	timeout_add(&wan_timer->timer_info, delay);
#else
# error "wan_add_timer() function is not supported yet!"
#endif /* linux */
}

/********************** WANPIPE KERNEL BUFFER **************************/
/*
** wan_skb_data() - 
**		Returns pointer to data.
*/
static __inline unsigned char* wan_skb_data(void* skb)
{
#if defined(__LINUX__)
	return ((struct sk_buff*)skb)->data;
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	return mtod((struct mbuf*)skb, caddr_t);
#else
# error "wan_skb_data() function is not supported yet!"
#endif
}

/*
** wan_skb_tail() - 
**		Returns pointer to data.
*/
static __inline unsigned char* wan_skb_tail(void* skb)
{
#if defined(__LINUX__)
	return ((struct sk_buff*)skb)->tail;
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	return mtod((struct mbuf*)skb, caddr_t);
#else
# error "wan_skb_data() function is not supported yet!"
#endif
}


/*
** wan_skb_len() - 
**		Returns current kernel buffer length.
*/
static __inline int wan_skb_len(void* skb)
{
#if defined(__LINUX__)
	return ((struct sk_buff*)skb)->len;
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	return ((struct mbuf*)skb)->m_len;
#else
# error "wan_skb_len() function is not supported yet!"
#endif
}


/*
** wan_skb_alloc() - 
**		Allocate kernel buffer with len.
*/
static __inline void* wan_skb_alloc(unsigned int len, void* dev)
{
#if defined(__LINUX__)
#if defined(WAN_DEBUG_MEM)
	struct sk_buff *skb=dev_alloc_skb(len);
	if (skb){
		DEBUG_ADD_MEM(skb->truesize);
	}
	return (void*)skb;
#else
	return (void*)dev_alloc_skb(len);
#endif
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	struct mbuf*	new_mbuf = NULL;
	MGETHDR(new_mbuf, M_DONTWAIT, MT_DATA);
	if (new_mbuf == NULL)
		return NULL;
	if (dev){
		new_mbuf->m_pkthdr.rcvif	= dev;
	}
	new_mbuf->m_pkthdr.len		= new_mbuf->m_len = 0;
	if ((len+2) > MHLEN){
		MCLGET(new_mbuf, M_DONTWAIT);
		if ((new_mbuf->m_flags & M_EXT) == 0){
			m_freem(new_mbuf);
			return NULL;
		}
	}
	new_mbuf->m_data += 2;
	return (void*)new_mbuf;
#else
# error "wan_skb_alloc() function is not supported yet!"
#endif
}

/*
** wan_skb_free() - 
**		Free kernel memory buffer.
*/
static __inline void wan_skb_free(void* skb)
{
#if defined(__LINUX__)
#if defined(WAN_DEBUG_MEM)
	DEBUG_SUB_MEM(((struct sk_buff*)skb)->truesize);
#endif
	wan_dev_kfree_skb(skb, FREE_WRITE);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	m_freem(skb);
#else
# error "wan_skb_free() function is not supported yet!"
#endif
}

/*
** wan_skb_check() - 
**		Check if packet consists from one skb block.
*/
static __inline int wan_skb_check(void* skb)
{
#if defined(__LINUX__)
	return 0;
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	netskb_t*	m = (netskb_t*)skb;
	if (m->m_pkthdr.len != m->m_len){
		return 1;
	}
	return 0;
#else
# error "wan_skb_check() function is not supported yet!"
#endif
}

/*
** wan_skb_reserve() - 
**		Adjust headroom.
*/
static __inline void wan_skb_reserve(void* skb, unsigned int len)
{
#if defined(__LINUX__)
	skb_reserve(skb, len);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
#else
# error "wan_skb_free() function is not supported yet!"
#endif
}

/*
** wan_skb_copyback() - 
** 	Copy data from a buffer back into the indicated mbuf chain,
** 	starting "off" bytes from the beginning, extending the mbuf
** 	chain if necessary.
*/
static __inline void wan_skb_copyback(void* skb, int off, int len, caddr_t cp)
{
#if defined(__LINUX__)
	struct sk_buff* sk = (struct sk_buff*)skb;
	unsigned char* data = NULL;
	if (off == wan_skb_len(skb)){
		if (sk->tail + len > sk->end){	
			DEBUG_EVENT("wan_skb_copyback: Internal Error (off=%d,len=%d,skb_len=%d)!\n",
					off, len, wan_skb_len(skb));
			return;
		}else{
			data = skb_put(skb, len);
			memcpy(data, cp, len);
		}
	}else{
		if (off + len > wan_skb_len(skb)){
			data = skb_put(skb, len);
			memcpy(data + off, cp, len);
			skb_trim(skb, off + len);
		}
	}
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	struct mbuf*	m = (struct mbuf*)skb;
	caddr_t		data = mtod(m, caddr_t);

	bcopy(cp, &data[off], len); 
	m->m_len 	= off + len;
	m->m_pkthdr.len = off + len;
#else
# error "wan_skb_copyback() function is not supported yet!"
#endif
}

/*
** wan_skb_copyback_user() - 
** 	Copy data from a buffer back into the indicated mbuf chain,
** 	starting "off" bytes from the beginning, extending the mbuf
** 	chain if necessary.
**      Data being copied is coming from user space, thus we must
**      use a special function to copy it into kernel space.
*/
static __inline int wan_skb_copyback_user(void* skb, int off, int len, caddr_t cp)
{
#if defined(__LINUX__)
	struct sk_buff* sk = (struct sk_buff*)skb;
	unsigned char* data = NULL;
	if (off == wan_skb_len(skb)){
		if (sk->tail + len > sk->end){	
			DEBUG_EVENT("wan_skb_copyback_user: Internal Error (off=%d,len=%d,skb_len=%d)!\n",
					off, len, wan_skb_len(skb));
			return -EINVAL;
		}else{
			data = skb_put(skb, len);
			if (WAN_COPY_FROM_USER(data, cp, len)){
				DEBUG_EVENT("wan_skb_copyback_user: Internal Error (off=%d,len=%d,skb_len=%d)!\n",
					off, len, wan_skb_len(skb));
				return -EFAULT;
			}
		}
	}else{
		if (off + len > wan_skb_len(skb)){
			data = skb_put(skb, len);
			if (WAN_COPY_FROM_USER(data+off, cp, len)){
				DEBUG_EVENT("wan_skb_copyback_user: Internal Error (off=%d,len=%d,skb_len=%d)!\n",
					off, len, wan_skb_len(skb));
				return -EFAULT;
			}
			skb_trim(skb, off + len);
		}
	}
	return 0;
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	struct mbuf*	m = (struct mbuf*)skb;
	caddr_t		data = mtod(m, caddr_t);

	WAN_COPY_FROM_USER(cp, &data[off], len);
	m->m_len 	= off + len;
	m->m_pkthdr.len = off + len;
#else
# error "wan_skb_copyback_user() function is not supported yet!"
#endif
	return 0;
}


/*
** wan_skb_copyback() - 
**	Copy data from an mbuf chain starting "off" bytes from the beginning,
**	continuing for "len" bytes, into the indicated buffer.
*/
static __inline void wan_skb_copydata(void* skb, int off, int len, caddr_t cp)
{
#if defined(__LINUX__)
	if (off + len > wan_skb_len(skb)){
		DEBUG_EVENT("wan_skb_copydata: Internal error (off=%d, len=%d, skb_len=%d)!\n",
					off, len, wan_skb_len(skb));
		return;
	}
	memcpy(cp, wan_skb_data(skb), len);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	caddr_t		data = mtod((struct mbuf*)skb, caddr_t);

	bcopy(cp, &data[off], len); 
#else
# error "wan_skb_copydata() function is not supported yet!"
#endif
}

/*
** wan_skb_correct() - 
**		Correct skb block.
*/
static __inline void wan_skb_correct(void* skb_dst, void* skb_src)
{
#if defined(__LINUX__)
	return;
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	netskb_t*	m_dst = (netskb_t*)skb_dst;
	netskb_t*	m_src = (netskb_t*)skb_src;
	if (!m_dst || !m_src){
		return;
	}
	while(m_src != NULL){
		wan_skb_copyback(m_dst, wan_skb_len(m_dst),
				 wan_skb_len(m_src), wan_skb_data(m_src));
		m_src = m_src->m_next;
	}
	return;
#else
# error "wan_skb_correct() function is not supported yet!"
#endif
}

/*
** wan_skb_pull() - 
**
*/
static __inline unsigned char* wan_skb_pull(void* skb, int len)
{
#if defined(__LINUX__)
	return skb_pull((struct sk_buff*)skb, len);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	m_adj(skb, len);
#if 0
	struct mbuf*	m = (struct mbuf*)skb;
	m->m_data += len;
	m->m_pkthdr.len -= len;
	m->m_len 	= m->m_pkthdr.len;
#endif
	return wan_skb_data(skb);
#else
# error "wan_skb_pull() function is not supported yet!"
#endif
}

/*
** wan_skb_put() - 
**
*/
static __inline unsigned char* wan_skb_put(void* skb, int len)
{
#if defined(__LINUX__)
	return skb_put((struct sk_buff*)skb, len);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	struct mbuf*	m = (struct mbuf*)skb;
	void*		tail = wan_skb_data(skb) + m->m_len;

	m->m_len 	= wan_skb_len(skb) + len;
	m->m_pkthdr.len = wan_skb_len(skb);
	/* Original before Mar 19,2004 return wan_skb_data(skb);*/
	return (char*)tail;
#else
# error "wan_skb_put() function is not supported yet!"
#endif
}

/*
** wan_skb_push() - 
**
*/
static __inline unsigned char* wan_skb_push(void* skb, int len)
{
#if defined(__LINUX__)
	return skb_push((struct sk_buff*)skb, len);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	/* wan_skb_push() function is not supported yet! */
	return wan_skb_data(skb);
#else
# error "wan_skb_push() function is not supported yet!"
#endif
}




/*
** wan_skb_tailroom() - Tail room
**			
** 
*/
static __inline int wan_skb_tailroom(void* skb)
{
#if defined(__LINUX__)
	return skb_tailroom(skb);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	struct mbuf*	m = (struct mbuf*)skb;

	if (m->m_flags & M_EXT){
		return (MCLBYTES - m->m_len);
	}
	return (MHLEN - m->m_len);
#else
# error "wan_skb_tailroom() function is not supported yet!"
#endif
}

/*
** wan_skb_tailroom() - Head room
**			
** 
*/
static __inline int wan_skb_headroom(void* skb)
{
#if defined(__LINUX__)
	return skb_headroom(skb);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	return 0;
#else
# error "wan_skb_headroom() function is not supported yet!"
#endif
}



/*
** wan_skb_trim() - Trim from tail
**			
** 
*/
static __inline void wan_skb_trim(void* skb, unsigned int len)
{
#if defined(__LINUX__)
	skb_trim(skb, len);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	struct mbuf*	m = (struct mbuf*)skb;
	if (len == 0){
		m->m_data 	= m->m_ext.ext_buf;
	}
	m->m_pkthdr.len	= len;
	m->m_len	= m->m_pkthdr.len;
#else
# error "wan_skb_trim() function is not supported yet!"
#endif
}

/*
** wan_skb_init() - Setup skb data ptr
**			
** 
*/
static __inline void wan_skb_init(void* pskb, unsigned int len)
{
#if defined(__LINUX__)
	struct sk_buff* skb = (struct sk_buff*)pskb;
	skb->data = skb->head + len;
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
#else
# error "wan_skb_init() function is not supported yet!"
#endif
}


/*
** wan_skb_queue_init() - 
**
*/
static __inline void wan_skb_queue_init(void *list)
{
#if defined(__LINUX__)
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
#else
# error "wan_skb_queue_init() function is not supported yet!"
#endif
}

/*
** wan_skb_queue_purge() - 
**
*/

static __inline void wan_skb_queue_purge(void *list)
{
#if defined(__LINUX__)
	struct sk_buff *skb;
	while ((skb=skb_dequeue(list))!=NULL){
		wan_skb_free(skb);
	}
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	IFQ_PURGE(((wan_skb_queue_t*)list));
#else
# error "wan_skb_queue_purge() function is not supported yet!"
#endif
}

/*
** wan_skb_queue_tail() - 
**
*/
static __inline void wan_skb_queue_tail(void* list, void* newsk)
{
#if defined(__LINUX__)
	skb_queue_tail(list, newsk);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	IF_ENQUEUE((wan_skb_queue_t*)list, (struct mbuf*)newsk);
#else
# error "wan_skb_queue_tail() function is not supported yet!"
#endif
}

/*
** wan_skb_queue_head() - 
**
*/
static __inline void wan_skb_queue_head(void* list, void* newsk)
{
#if defined(__LINUX__)
	skb_queue_head(list, newsk);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	IF_PREPEND((wan_skb_queue_t*)list, (struct mbuf*)newsk);
#endif
}


/*
** wan_skb_queue_len() - 
**
*/
static __inline int wan_skb_queue_len(void* list)
{
#if defined(__LINUX__)
	return skb_queue_len(list);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	wan_skb_queue_t* ifqueue = (wan_skb_queue_t*)list;
	return IFQ_LEN(ifqueue);
#else
# error "wan_skb_queue_len() function is not supported yet!"
#endif
}


/*
** wan_skb_dequeue() - 
**
*/
static __inline void *wan_skb_dequeue(void* list)
{
#if defined(__LINUX__)
	return skb_dequeue(list);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	netskb_t*	newsk = NULL;
	
	IF_DEQUEUE((wan_skb_queue_t*)list, newsk);
	return (void*)newsk;
#else
# error "wan_skb_dequeue() function is not supported yet!"
#endif

}


/*
**************************************************************************
**			 WANPIPE NETWORK INTERFACE 			**
**************************************************************************
*/

static __inline void* wan_netif_alloc(unsigned char *devname,int *err)
{
#if defined(__LINUX__)
# if defined(LINUX_2_6)
	return __dev_alloc(devname, err);
# elif defined(LINUX_2_4)
	netdevice_t *dev=wan_malloc(sizeof(netdevice_t));
	*err=0;

	if (!dev){
		*err=-ENOMEM;
	}
	memset(dev, 0, sizeof(netdevice_t));
	
	strncpy(dev->name,devname,30);
	return dev;	
# else	
	netdevice_t *dev=wan_malloc(sizeof(netdevice_t));
	*err=0;

	if (!dev){
		*err=-ENOMEM;
	}
	memset(dev, 0, sizeof(netdevice_t));

	dev->name=wan_malloc(35);
	if (!dev->name){
		wan_free(dev);
		dev=NULL;
		*err=-ENOMEM;
		return NULL;
	}
	strncpy(dev->name,devname,30);
	return dev;
# endif
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	struct ifnet* ifp;
	ifp = wan_malloc(sizeof(struct ifnet));
	if (ifp == NULL){
		*err = -ENOMEM;
		return NULL;
	}
	*err = wan_netif_init(ifp, devname, 0);
	if (*err){
		wan_netif_del(ifp);
		return NULL;
	};
	return ifp;
#else
# error "wan_netif_alloc() unsupported"
#endif

}

static __inline void wan_netif_free(netdevice_t *dev)
{

#if defined(__LINUX__)	
# if defined(LINUX_2_6)
	free_netdev(dev);
# elif defined(LINUX_2_4)
	wan_free(dev);
# else
	if (dev->name){
		wan_free(dev->name);
		dev->name=NULL;
	}
	wan_free(dev);
# endif
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	wan_netif_del(dev);
	wan_free(dev);
#else
#error "wan_netif_free() not supported!"
#endif

}

	

static __inline int wan_netif_init(netdevice_t* dev, char* ifname, int ifunit)
{
#if defined(__FreeBSD__)
	int	len = 0;
        while((ifname[len] != '\0') && (len <= strlen(ifname))){
		if ((ifname[len] >= '0') && (ifname[len] <= '9')){
			break;
		}
		len++;    
	}
	dev->if_unit	= ifunit;
	if (dev->if_name == NULL){
		dev->if_name = malloc(len+1, M_DEVBUF, M_NOWAIT);
	}
	if (dev->if_name == NULL){
		return -ENOMEM;
	}
	memset(dev->if_name, 0, len+1);
	bcopy(ifname, dev->if_name, len);
	dev->if_name[len] = '\0';
	IFQ_SET_MAXLEN(&dev->if_snd, ifqmaxlen);
#elif defined(__OpenBSD__)
	if (strlen(ifname) >= IFNAMSIZ){
		return -ENOMEM;
	}
    	bcopy(ifname, dev->if_xname, strlen(ifname));
	IFQ_SET_MAXLEN(&dev->if_snd, IFQ_MAXLEN);
#elif defined(__LINUX__)
# if  (LINUX_VERSION_CODE < KERNEL_VERSION(2,3,43))
	dev->name = ifname;
# else	
	strcpy(dev->name, ifname);
# endif
#else
# error "wan_netif_init() function is not supported yet!"
#endif
	return 0;
}


static __inline int wan_netif_del(netdevice_t* dev)
{
	WAN_ASSERT(dev == NULL);
#if defined(__FreeBSD__)
	dev->if_init = NULL;
# if (__FreeBSD_version > 400000)
	/* Free interface name (only for FreeBSD-4.0) */
	free(dev->if_name, M_DEVBUF);
# endif
#elif defined(__OpenBSD__)
	/* Do nothing */
#elif defined(__LINUX__)
	/* Do nothing */
#else
# error "wan_netif_del() function is not supported yet!"
#endif
	return 0;
}

static __inline char* wan_netif_name(netdevice_t* dev)
{
	static char ifname[IFNAMSIZ+1];
	WAN_ASSERT2(dev == NULL, NULL);
#if defined(__LINUX__)
	strcpy(ifname, dev->name);
#elif defined(__FreeBSD__)
	sprintf(ifname, "%s%d", dev->if_name, dev->if_unit);
#elif defined(__OpenBSD__)
	sprintf(ifname, "%s", dev->if_xname);
#else
# error "wan_get_ifname() function is not supported yet!"
#endif
	return ifname;
}

static __inline void* wan_netif_priv(netdevice_t* dev)
{
	WAN_ASSERT2(dev == NULL, NULL);
#if defined(__LINUX__)
	return dev->priv;
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	return dev->if_softc;
#else
# error "wan_netif_priv() function is not supported yet!"
#endif
}

static __inline void wan_netif_set_priv(netdevice_t* dev, void* priv)
{
	WAN_ASSERT1(dev == NULL);
#if defined(__LINUX__)
	dev->priv = priv;
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	dev->if_softc = priv;
#else
# error "wan_netif_priv() function is not supported yet!"
#endif
	return;
}

static __inline short wan_netif_flags(netdevice_t* dev)
{
	WAN_ASSERT(dev == NULL);
#if defined(__LINUX__)
	return dev->flags;
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	return dev->if_flags;
#else
# error "wan_netif_flags() function is not supported yet!"
#endif
}

static __inline void wan_netif_set_flags(netdevice_t* dev, short if_flags)
{
	WAN_ASSERT1(dev == NULL);
#if defined(__LINUX__)
	dev->flags = if_flags;
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	dev->if_flags = if_flags;
#else
# error "wan_netif_flags() function is not supported yet!"
#endif
}

static __inline int wan_netif_mcount(netdevice_t* dev)
{
#if defined(__LINUX__)
	return dev->mc_count;
#elif defined(__FreeBSD__)
	return dev->if_amcount;
#elif defined(__OpenBSD__)
	return 0;
#else
# error "wan_netif_mcount() function is not supported yet!"
#endif
}


static __inline void wan_bpf_report(netdevice_t* dev, void* pkt, int flag)
{

#if defined(__LINUX__)
	/* Do nothing */
#elif defined(__FreeBSD__)
	if (dev->if_bpf != NULL){ /* BF-0002 */ 
		WAN_BPF_REPORT(dev, pkt);
	}
#elif defined(__OpenBSD__)
	if (dev->if_bpf != NULL){ /* BF-0002 */ 
		if (flag){
			struct mbuf m0;
			u_int32_t af = AF_INET;
			m0.m_next = pkt;
			m0.m_len = 4;
			m0.m_data = (char*)&af;
			bpf_mtap(dev->if_bpf, &m0);
			WAN_BPF_REPORT(dev, &m0);
		}else{
			WAN_BPF_REPORT(dev, pkt);
		}
	}
#else
# error "wan_bpf_report() function is not supported yet!"
#endif

}


static __inline void wan_spin_lock_irq(void *lock, unsigned long *flag)
{
#if defined(__LINUX__)
	spin_lock_irqsave(((spinlock_t*)lock),*flag);	
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	*((int*)lock) = splimp();
#else
# warning "wan_spin_lock_irq() function is not supported yet!"
#endif	
}
static __inline void wan_spin_unlock_irq(void *lock, unsigned long *flag)
{
#if defined(__LINUX__)
	spin_unlock_irqrestore(((spinlock_t*)lock),*flag);	
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	splx(*(int*)lock);
#else
# warning "wan_spin_unlock_irq() function is not supported yet!"
#endif	
}

static __inline void wan_spin_lock(void *lock)
{
#if defined(__LINUX__)
	spin_lock(((spinlock_t*)lock));	
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	*((int*)lock) = splimp();
#else
# warning "wan_spin_lock() function is not supported yet!"
#endif	
}
static __inline void wan_spin_unlock(void *lock)
{
#if defined(__LINUX__)
	spin_unlock(((spinlock_t*)lock));	
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	splx(*(int*)lock);
#else
# warning "wan_spin_unlock() function is not supported yet!"
#endif	
}

#if 0
static __inline void wan_read_rw_lock(void *lock)
{
#if defined(__LINUX__)
	read_lock(((rwlock_t*)lock));
#else
# warning "wan_read_rw_lock() function is not supported yet!"
#endif
}

static __inline void wan_read_rw_unlock(void *lock)
{
#if defined(__LINUX__)
	read_unlock(((rwlock_t*)lock));
#else
# warning "wan_read_rw_unlock() function is not supported yet!"
#endif
}

static __inline void wan_write_rw_lock_irq(void *lock, unsigned long *flag)
{
#if defined(__LINUX__)
	write_lock_irqsave(((rwlock_t*)lock),flag);
#else
# warning "wan_read_rw_lock() function is not supported yet!"
#endif
}

static __inline void wan_write_rw_unlock_irq(void *lock, unsigned long *flag)
{
#if defined(__LINUX__)
	write_unlock_irqrestore(((rwlock_t*)lock),flag);
#else
# warning "wan_read_rw_unlock() function is not supported yet!"
#endif
}
#endif

#endif	/* __WANPIPE_COMMON_H */
