/*
 * Copyright (c) 1997, 1998, 1999
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: sdla_dev.c.FreeBSD,v 1.11 2003/07/31 15:25:45 sangoma Exp $
 */

/*****************************************************************************
 * sdla_dev.c	SDLA DEV Support Module.  Main module.
 *
 *		This module is a compeled with kernel sources.
 *
 * Author:	Alex Feldman	<al.feldman@sangoma.com>
 ****************************************************************************
 */

#undef INTR_TEST	/* uncomment this to check that driver RX INTR */

#define WAN_KERNEL

#include <sys/types.h>
#include <sys/param.h>
#if (__FreeBSD_version < 500000)
# include "sdla.h"		/* number of Sangoma cards */
# include <stddef.h>
#else
# define NSDLA		20	/* temporary value */
#endif
#include <sys/systm.h>
#include <sys/syslog.h>
#include <sys/errno.h>
#include <sys/mbuf.h>
#include <sys/proc.h>
#include <sys/sockio.h>
#include <sys/socket.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/sysent.h>
#if (__FreeBSD_version > 400000)
# include <sys/bus.h>
#endif
#include <sys/interrupt.h>
#include <machine/bus.h>
#if (__FreeBSD_version > 400000)
# include <sys/rman.h>
# include <machine/resource.h>
#endif

#include <net/wanpipe_debug.h>
#include <net/sdlapci.h>
#include <net/sdlasfm.h>
#include <net/sdla_front_end.h>
#include <net/sdla_dev.h>	/* special header for S508 board */

#include <i386/isa/isa_device.h>

#include <vm/vm.h>
#include <vm/pmap.h>
 
#include <pci/pcivar.h>
#include <pci/pcireg.h>

/* hardware related constant */
#define MAX_CARDS         8
#define S508_NUMPORTS     2
#define S508_ID_BYTE      0x11
#define S508_IO_PORTS	  4

/* SDLA ISA global constants */
#define MOD_VERSION     1
#define MOD_RELEASE     1
#define MOD_MINOR       0

/* Maximum number cards (each CPU like separate card ) */
#define MAX_SDLA_CARD 	NSDLA*2

#if (__FreeBSD_version <= 400000)
# device_get_name(dev)	SDLA_DEV_NAME	
#endif

/****** Global Data *******************************************************/
int    	Sangoma_cards_no = 0;
int    	Sangoma_devices_no = 0;
int	Sangoma_PCI_cards_no = 0; 
sdladev_t sdladev[MAX_SDLA_CARD];

/*
********************************************************************
****								****
****								****
****	        	    FreeBSD-4.x				****
****								****
****								****
********************************************************************
*/
#if (__FreeBSD_version > 400000)

/**************************************************************************/
/* SDLA ISA device related entry point 		  			  */
static int 	sdla_isa_probe __P((device_t));
static int 	sdla_isa_attach __P((device_t));
driver_intr_t	sdla_isa_intr;

/* isa_driver structure for autoconfiguration process */
static device_method_t sdla_isa_methods[] = {
	DEVMETHOD(device_probe,		sdla_isa_probe),
	DEVMETHOD(device_attach,	sdla_isa_attach),
	{0, 0}
};

static driver_t	sdla_isa_driver = {
	SDLA_DEV_NAME,
	sdla_isa_methods,
	sizeof(struct sdla_softc)
};

static devclass_t sdla_isa_devclass;

DRIVER_MODULE(sdla, isa, sdla_isa_driver, sdla_isa_devclass, 0, 0);

/*************************************************************************/
/* SDLA PCI device relative entry point 	 	 		 */

static int	sdla_pci_probe	__P((device_t));
static int	sdla_pci_attach	__P((device_t));
driver_intr_t	sdla_pci_intr;

static device_method_t sdla_pci_methods[] = {
	DEVMETHOD(device_probe,		sdla_pci_probe),
	DEVMETHOD(device_attach,	sdla_pci_attach),
	{0, 0}
};

static driver_t	sdla_pci_driver = {
	SDLA_DEV_NAME,
	sdla_pci_methods,
	sizeof(struct sdla_softc),
};

static devclass_t sdla_devclass;

DRIVER_MODULE(sdla, pci, sdla_pci_driver, sdla_devclass, 0, 0);
/***********************************************************************/
int sdla_alloc_port(device_t dev, int size);
int sdla_alloc_memory(device_t dev, int size);
int sdla_alloc_irq(device_t dev, int flags);
void sdla_release_resources(device_t dev);
/***********************************************************************/

/******* ISA Device Driver Entry Points ********************************/
/*
 ************************************************************************
 * sdla_isa_probe() - called during system's autoconfiguration to probe for
 * 		     device presence. The function should return the number
 * 		     of i/o ports needed, or 0 if device not present.
 */
static int
sdla_isa_probe(device_t dev)
{
    	struct sdla_softc*	sc = device_get_softc(dev);
    	sdladev_t*		isa_cfg = NULL;
    	u_long			junk, irq;
    	int			error = 0;
    	unsigned char	id_byte = 0x0;

    	bzero(sc, sizeof(struct sdla_softc));

  	/* Check that number of cards not bigger than size of array (MAX_SDLA_CARD) */
    	if (Sangoma_cards_no >= MAX_SDLA_CARD){
		DEBUG_EVENT("%s: Exceed muximum number of Sangoma cards (%d)!\n",
				device_get_name(dev), Sangoma_cards_no);
		return ENOENT; 
    	}
    	/* Grab some info */
    	isa_cfg 	= &sdladev[Sangoma_cards_no];
    	bzero(isa_cfg, sizeof(sdladev_t));
    	isa_cfg->type  	= SDLA_S508;
    	isa_cfg->sc    	= sc;
    	isa_cfg->isr   	= NULL;
    	sc->dev		= dev;
#if (__FreeBSD_version >= 450000)
    	isa_cfg->dev	= dev;
#endif
 
    	/* Try see if device is these. Read iobase+3and check if se see
     	 * 0x11 which S508's identifier.
     	 */
    	error = sdla_alloc_port(dev, S508_IO_PORTS);
    	if (error){
		DEBUG_EVENT("%s: Failed allocate I/O port!\n",
					device_get_name(dev));
		goto sdla_isa_probe_done;
    	}
    	isa_cfg->u.isa.ioport = rman_get_start(sc->port_res);
    	id_byte = s508_read_port(isa_cfg->u.isa.ioport, SDLA_S508_OFFSET);
    	if (id_byte != S508_ID_BYTE){
		DEBUG_KERNEL("%s: This is not Sangoma card (id_byte = %02x)!\n",
					device_get_name(dev), id_byte);
		error = ENOENT;
		goto sdla_isa_probe_done;
    	}

    	error = sdla_alloc_memory(dev, 0x2000);
    	if (error){
		DEBUG_EVENT("%s: Failed allocate memory!\n",
					device_get_name(dev));
		goto sdla_isa_probe_done;
    	}
    	isa_cfg->u.isa.maddr = rman_get_virtual(sc->mem_res);

    	error = bus_get_resource(dev, SYS_RES_IRQ, 0, &irq, &junk);
    	if (error){
		DEBUG_EVENT("%s: Failed get IRQ interrupt!\n",
					device_get_name(dev));
		goto sdla_isa_probe_done;
    	}
    	isa_cfg->irq = (int)irq;

    	if ((error = sdla_alloc_irq(dev, 0))){ 
		DEBUG_EVENT("%s: IRQ %d is already used by another device!\n",
					device_get_name(dev), (int)irq);
		goto sdla_isa_probe_done;
    	}
sdla_isa_probe_done:
    	sdla_release_resources(dev);
    
    	/* OK, device is there, return number of ports */
    	return error; 	
}

/*
 **************************************************************************
 * sdla_isa_attach() - called if the probe is succeeded. Attaches wanpipe?
 *                    device to the stack.
 */
static int
sdla_isa_attach(device_t dev)
{
    	struct sdla_softc* 	sc = device_get_softc(dev);
    	sdladev_t*		isa_cfg = &sdladev[Sangoma_cards_no];
    	int			error = 0;

    	isa_cfg->isr  = NULL;
    	isa_cfg->arg = NULL;
    	if (sc->port_used){
		sdla_alloc_port(dev, sc->port_used);
    	}
    	if (sc->mem_used){
		sdla_alloc_memory(dev, sc->mem_used);
    	}
    	if ((error = sdla_alloc_irq(dev, 0))){
		DEBUG_EVENT("%s: IRQ %d is already used by another device!\n",
				device_get_name(dev), isa_cfg->irq);
		return error;
    	}
#if (__FreeBSD_version < 450000)
    	error = bus_setup_intr(dev, 
			       sc->irq_res, 
			       INTR_TYPE_NET, 
			       sdla_isa_intr, 
			       isa_cfg, 
			       &isa_cfg->irqh);
    	if (error){
		sdla_release_resources(dev);
		DEBUG_EVENT("%s: Failed set interrupt handler!\n",
					device_get_name(dev));
		return error;
    	} 
#endif
    	/* increment number of Sangoma adapters */
    	Sangoma_cards_no++;
    	/* increment number of Sangoma devices */
    	Sangoma_devices_no += 2;
    	return 0;
}

/*
 ****************************************************************************
 * sdla_isa_intr() - Dummy ISA device interrupt handler.
 */
void sdla_isa_intr(void* arg)
{
    	sdladev_t* 		isa_cfg = (sdladev_t*)arg;
    	struct sdla_softc*	sc = NULL;
 
#if (__FreeBSD_version >= 450000)
	DEBUG_EVENT("sdla_isa_intr: should never happened!\n");
#endif
	DEBUG_KERNEL("%s?: Interrupt from S508 Adapter!\n",
					SDLA_DEV_NAME);
	if (isa_cfg == NULL || 
	    isa_cfg->sc == NULL ||
	    isa_cfg->type != SDLA_S508){
		DEBUG_EVENT("%s?: Uncorrect adapter type!\n",
					SDLA_DEV_NAME);
		return;
    	}

    	sc = isa_cfg->sc;
    	if (isa_cfg->isr && (isa_cfg->arg != NULL)){
		isa_cfg->isr(isa_cfg->arg);
	}
    	return;
}


/******* PCI Device Driver Entry Points ********************************/
static int sdla_pci_probe(device_t dev)
{
    	u_int16_t	vendor = pci_get_vendor(dev);
    	u_int16_t	subvendor = pci_get_subvendor(dev);
    	u_int16_t	device = pci_get_device(dev);
    	u_int16_t	adapter_type = 0;

    	/* Check ivendor and subvendor ID */
    	if (!((vendor == V3_VENDOR_ID && subvendor == SANGOMA_SUBSYS_VENDOR) ||
    	      (vendor == PCI_VENDOR_ID_GSI && device == PCI_DEVICE_ID_GSI_PULSAR))){
		DEBUG_KERNEL("%s: not our card (vendor=%04x, subvendor=%04x)!\n",
					device_get_name(dev),vendor,subvendor);
		return ENXIO;
    	}

    	/* Check that number of cards not bigger than size of array 
	 * (MAX_SDLA_CARD) 
	 */
    	if (Sangoma_cards_no >= MAX_SDLA_CARD){
		DEBUG_EVENT("%s: Exceed muximum number of Sangoma cards (%d)!\n",
				device_get_name(dev), Sangoma_cards_no);
		return ENXIO;
    	}
    	if (Sangoma_PCI_cards_no >= MAX_S514_CARDS){
		DEBUG_EVENT("%s: Exceed muximum number of Sangoma PCI cards (%d)!\n",
				device_get_name(dev), Sangoma_PCI_cards_no);
		return ENXIO;
    	}

	adapter_type = pci_read_config(dev, PCI_SUBSYS_ID_WORD, 2);
	switch(vendor){
	case V3_VENDOR_ID:
		switch(adapter_type){
		case S5141_ADPTR_1_CPU_SERIAL:
    			device_set_desc(dev, "Sangoma S5141/FT1 (Single CPU) adapter");
			break;
		case S5142_ADPTR_2_CPU_SERIAL:
    			device_set_desc(dev, "Sangoma S5142 (Dual CPU) adapter");
			break;
		case S5143_ADPTR_1_CPU_FT1:
    			device_set_desc(dev, "Sangoma S5143 (Single CPU) FT1 adapter");
			break;
		case S5144_ADPTR_1_CPU_T1E1:
		case S5148_ADPTR_1_CPU_T1E1:
    			device_set_desc(dev, "Sangoma S5144 (Single CPU) T1/E1 adapter");
			break;
		case S5145_ADPTR_1_CPU_56K:		
    			device_set_desc(dev, "Sangoma S5145 (Single CPU) 56K adapter");
			break;
		case S5147_ADPTR_2_CPU_T1E1:
    			device_set_desc(dev, "Sangoma S5147 (Dual CPU) T1/E1 adapter");
			break;
		}
		break;

	case PCI_VENDOR_ID_GSI:
    		device_set_desc(dev, "Sangoma S518 ADSL adapter");
		break;
	}	
    	return 0;
}

static int sdla_pci_attach(device_t dev)
{
    	struct sdla_softc*	sc = device_get_softc(dev);
    	sdladev_t*		pci_cfg = NULL;
    	sdladev_t*		pci_cfg_nxt = NULL;
    	u_int16_t		vendor = pci_get_vendor(dev);
    	u_int16_t		subvendor = pci_get_subvendor(dev);
    	u_int16_t		device = pci_get_device(dev);
	u_int16_t		adapter_type = pci_read_config(dev, PCI_SUBSYS_ID_WORD, 2);
    	int			type = 0, dual_cpu = 0, error = 0;

    	/* Check ivendor and subvendor ID */
    	if (vendor == V3_VENDOR_ID && subvendor == SANGOMA_SUBSYS_VENDOR){
		type = SDLA_S514;
		if (adapter_type == S5148_ADPTR_1_CPU_T1E1){
			adapter_type = S5144_ADPTR_1_CPU_T1E1;
		}
    	}else if (vendor == PCI_VENDOR_ID_GSI && device == PCI_DEVICE_ID_GSI_PULSAR){
		type = SDLA_ADSL;
		adapter_type = S518_ADPTR_1_CPU_ADSL;
    	}else{
		DEBUG_KERNEL("%s: not our card (vendow=%04x, subvendor=%04x)!\n",
					device_get_name(dev),vendor,subvendor);
		return ENXIO;
    	}

    	if (Sangoma_cards_no+1 >= MAX_SDLA_CARD){
		DEBUG_EVENT("%s: Exceed muximum number of Sangoma cards (%d)!\n",
				device_get_name(dev), Sangoma_cards_no);
		return ENXIO;
    	}
    	/* Check number of CPUs */
	switch(adapter_type){
	case S5142_ADPTR_2_CPU_SERIAL:
	case S5147_ADPTR_2_CPU_T1E1:
		dual_cpu = 1;
		break;
	}

    	/* Initialize common member for both CPU's */
    	sc->dev          = dev;
    	/* Allocate irq for both CPU's */
    	error = sdla_alloc_irq(dev, RF_SHAREABLE);
    	if (error){
		DEBUG_EVENT("%s: Failed allocate shared IRQ for Sangoma PCI device!\n",
				device_get_name(dev));
		sdla_release_resources(dev);
		return (error);
    	}

    	/* Map interrupt handler for first CPU */
    	pci_cfg			= &sdladev[Sangoma_cards_no];
    	pci_cfg->type 		= type;
    	pci_cfg->adapter_type	= adapter_type;
    	pci_cfg->vendor_id	= vendor;
    	pci_cfg->device_id	= device;
    	pci_cfg->isr		= NULL;
    	pci_cfg->arg		= NULL;
    	pci_cfg->irq		= pci_get_irq(dev);
    	pci_cfg->u.pci.cpu_no	= SDLA_CPU_A;
    	pci_cfg->u.pci.slot_no	= pci_get_slot(dev);
    	pci_cfg->u.pci.bus_no	= pci_get_bus(dev);
    	pci_cfg->sc   		= sc;
#if (__FreeBSD_version < 450000)
    	error = bus_setup_intr(dev, sc->irq_res, INTR_TYPE_NET,
				sdla_pci_intr, pci_cfg, &pci_cfg->irqh);
    	if (error){
		DEBUG_EVENT("%s: Failed set interrupt handler for CPU A!\n",
					device_get_name(dev));
		sdla_release_resources(dev);
		return error;
    	}
#else
    	pci_cfg->dev   		= dev;
#endif

    	if (dual_cpu){
    		/* Map interrupt handler for second CPU */
    		pci_cfg_nxt			= &sdladev[Sangoma_cards_no+1];
    		pci_cfg_nxt->type 		= type;
    		pci_cfg_nxt->adapter_type	= adapter_type;
    		pci_cfg_nxt->vendor_id		= vendor;
    		pci_cfg_nxt->device_id		= device;
    		pci_cfg_nxt->isr		= NULL;
    		pci_cfg_nxt->arg		= NULL;
    		pci_cfg_nxt->irq		= pci_get_irq(dev);
    		pci_cfg_nxt->u.pci.cpu_no	= SDLA_CPU_B;
    		pci_cfg_nxt->u.pci.slot_no	= pci_get_slot(dev);
    		pci_cfg_nxt->u.pci.bus_no	= pci_get_bus(dev);
    		pci_cfg_nxt->sc 		= sc;
#if (__FreeBSD_version < 450000)
    		error = bus_setup_intr(dev, sc->irq_res, INTR_TYPE_NET,
				sdla_pci_intr, pci_cfg_nxt, &pci_cfg_nxt->irqh);
    		if (error){
			DEBUG_EVENT("%s: Failed set interrupt handler for CPU B!\n",
					device_get_name(dev));
			
			sdla_release_resources(dev);
			return error;
    		}
#else
    		pci_cfg_nxt->dev   		= dev;
#endif
    	}

    	/* Increment number of SDLA adapters, include CPU_A and CPU_B */
    	Sangoma_cards_no += (dual_cpu) ? 2 : 1;
    	/* Incremenet number of S514 cards */
    	Sangoma_PCI_cards_no++;
	/* Increment number of Sangoma devices */
	switch(adapter_type){
	case S5141_ADPTR_1_CPU_SERIAL:
		Sangoma_devices_no += 2;
		break;
	case S5142_ADPTR_2_CPU_SERIAL:
		Sangoma_devices_no += 4;
		break;
	case S5143_ADPTR_1_CPU_FT1:
	case S5144_ADPTR_1_CPU_T1E1:
	/*case S5148_ADPTR_1_CPU_T1E1: The same value as _ADSL */
	case S5145_ADPTR_1_CPU_56K:		
	case S518_ADPTR_1_CPU_ADSL: 
		Sangoma_devices_no ++;
		break;
	case S5147_ADPTR_2_CPU_T1E1:
		Sangoma_devices_no += 2;
		break;
	}

    	return 0;
}


/****************************************************************************/
/*		Common resource functions				    */
/****************************************************************************/
/*
 * Allocate a port resource with given resource id.
 */
int sdla_alloc_port(dev, size)
	device_t	dev;
	int		size;
{
    	struct sdla_softc*	sc = device_get_softc(dev);

    	sc->port_res = 
	   bus_alloc_resource(dev, SYS_RES_IOPORT, &sc->port_rid, 0ul, ~0ul, size, RF_ACTIVE);
	
    	if (!sc->port_res){
		sc->port_rid = 0x0;
		sc->port_used = 0x0;
		return ENOENT;
	}else{
		sc->port_used = size;
	}
	return 0;
}

/*
 * Allocate an irq resource with the given resource id.
 */
int sdla_alloc_irq(dev, flags)
	device_t	dev;
	int		flags;
{
    	struct sdla_softc*	sc = device_get_softc(dev);

    	sc->irq_res = 
		bus_alloc_resource(dev, SYS_RES_IRQ, &sc->irq_rid, 0ul, ~0ul, 1, (RF_ACTIVE | flags));
    	if (!sc->irq_res){
		sc->irq_rid = NULL;
		return ENOENT;		
	}
	return 0;
}

/*
 * Allocate a memory resource with the given resource id.
 */
int sdla_alloc_memory(dev, size)
	device_t	dev;
	int		size;
{
    	struct sdla_softc*	sc = device_get_softc(dev);

    	sc->mem_res = bus_alloc_resource(dev, SYS_RES_MEMORY, &sc->mem_rid, 0ul, ~0ul, size, RF_ACTIVE);
	if (!sc->mem_res){
		sc->mem_used = 0x0;
		sc->mem_rid = 0x0;
		return ENOENT;
	}else{
		sc->mem_used = size;
	}
	return 0;
}

/*
 * Release all resources
 */
void sdla_release_resources(dev)
	device_t	dev;
{
    	struct sdla_softc*  sc = device_get_softc(dev);

    	if (sc->irq_res){
		bus_release_resource(dev, SYS_RES_IRQ, sc->irq_rid, sc->irq_res);
		sc->irq_res = 0;
		sc->irq_rid = 0;
    	}
	if (sc->port_res){
		bus_release_resource(dev, SYS_RES_IOPORT, 
			     sc->port_rid, sc->port_res);
		sc->port_res = 0;
		sc->port_rid = 0;
    	}
	if (sc->mem_res){
		bus_release_resource(dev, SYS_RES_MEMORY, 
			     sc->mem_rid, sc->mem_res);
		sc->mem_res = 0;
		sc->mem_rid = 0;
    	}
	return;
}

/*
********************************************************************
****								****
****								****
****	        	    FreeBSD-3.x				****
****								****
****								****
********************************************************************
*/
#else

/* SDLA ISA device related entry point */
static int 	   sdla_isa_probe(struct isa_device *dev);
static int 	   sdla_isa_attach(struct isa_device *dev);
static ointhand2_t sdla_isa_intr;

/* isa_driver structure for autoconfiguration process */
struct isa_driver sdladriver = {
	sdla_isa_probe, sdla_isa_attach, SDLA_DEV_NAME
};

/* SDLA PCI device relative entry point */
static unsigned long sdla_pci_count = 0;
static const char    *sdla_pci_probe __P((pcici_t, pcidi_t));
static void 	     sdla_pci_attach __P((pcici_t, int));
static pci_inthand_t sdla_pci_intr;

static struct pci_device sdlapcidev = {
	SDLA_DEV_NAME,
	sdla_pci_probe,
	sdla_pci_attach,
	&sdla_pci_count,
	NULL,
};
DATA_SET(pcidevice_set, sdlapcidev);

/******* ISA Device Driver Entry Points ********************************/
/*
 *************************************************************************
 * sdla_isa_probe() - called during system's autoconfiguration to probe for
 * 		     device presence. The function should return the number
 * 		     of i/o ports needed, or 0 if device not present.
 */
static int
sdla_isa_probe(struct isa_device *dev)
{
    	int 		isa_unit = dev->id_unit;
	unsigned char	id_byte = 0x00;

    	/* Check that number of cards not bigger than size of array (MAX_SDLA_CARD) */
    	if (Sangoma_cards_no >= MAX_SDLA_CARD){
		DEBUG_EVENT("%s: Exceed muximum number of Sangoma cards (%d)!\n",
				dev->id_driver->name, Sangoma_cards_no);
		return 0; 
	}
    	/* Check if the unit makes sense. */ 
    	if (isa_unit >= MAX_CARDS){
		DEBUG_EVENT("%s: Exceed muximum number of Sangoma ISA cards (%d)!\n",
				dev->id_driver->name, isa_unit);
		return 0;
    	}
    	/* Grab some info */
    	sdladev[Sangoma_cards_no].type           = SDLA_S508;
    	sdladev[Sangoma_cards_no].isr            = NULL;
    	sdladev[Sangoma_cards_no].irq      	 = (int)(ffs(dev->id_irq) - 1);
    	sdladev[Sangoma_cards_no].u.isa.isa_unit = isa_unit;
    	sdladev[Sangoma_cards_no].u.isa.maddr    = dev->id_maddr; /* virtual address */
    	sdladev[Sangoma_cards_no].u.isa.ioport   = dev->id_iobase; 
    
    	/* Try see if device is these. Read iobase+3and check if se see
     	 * 0x11 which S508's identifier.
     	 */
    	if ((id_byte = s508_read_port(dev->id_iobase, 3)) != S508_ID_BYTE){
		DEBUG_KERNEL("%s: This is not Sangoma ISA card (id_byte = %02x)!\n",
					dev->isa_driver->name, id_byte);
        	return 0;
	}

    	/* OK, device is there, return number of ports */
    	return S508_NUMPORTS; 	
}

/*
 **************************************************************************
 * sdla_isa_attach() - called if the probe is succeeded. Attaches wanpipe?
 *                    device to the stack.
 */
static int
sdla_isa_attach(struct isa_device *dev)
{
    	if (Sangoma_cards_no >= MAX_SDLA_CARD){
		DEBUG_EVENT("%s: Exceed muximum number of Sangoma cards (%d)!\n",
				dev->id_driver->name, Sangoma_cards_no);
		return 0;
	}
    	dev->id_ointr = sdla_isa_intr;
    	sdladev[Sangoma_cards_no].isr = NULL;
    	sdladev[Sangoma_cards_no].arg = NULL;
    	/* increment number of Sangoma adapters */
    	Sangoma_cards_no++;
    	/* increment number of Sangoma devices */
    	Sangoma_devices_no += 2;
    	return 1;
}

/*
 ****************************************************************************
 * sdla_isa_intr() - Dummy ISA device interrupt handler.
 */
static void sdla_isa_intr(int isa_unit)
{
    	sdladev_t *isa_cfg = NULL;
    	int              i = 0;
 
	DEBUG_KERNEL("%s%d: Interrupt from S508 Adapter!!\n",
					SDLA_DEV_NAME, isa_unit);
    	if (isa_unit >= Sangoma_cards_no){
		DEBUG_EVENT("%s%d: Wrong Sangoma ISA card index!\n",
					SDLA_DEV_NAME, isa_unit);
		return;
	}
    	for(i = 0; (i < Sangoma_cards_no && isa_cfg == NULL); i++){
		if (sdladev[i].type != SDLA_S508)
	    		continue;
		if (sdladev[i].u.isa.isa_unit != isa_unit)
	    		continue;
		isa_cfg = &sdladev[i]; 
    	}

    	if (isa_cfg == NULL){
		DEBUG_EVENT("%s%d: Uncorrect addapter type!\n", 
					SDLA_DEV_NAME, isa_unit);
		return; 
    	}
    	if (isa_cfg->isr && (isa_cfg->arg != NULL)){
		isa_cfg->isr(isa_cfg->arg);
	}
    	return;
}


/******* PCI Device Driver Entry Points ********************************/

static const char *sdla_pci_probe(pcici_t config_id, pcidi_t device_id)
{
    	uint16_t PCI_vendor, PCI_device;
    	uint16_t PCI_subsys_vendor;

    	/* Check that number of cards not bigger than size of array (MAX_SDLA_CARD) */
    	if (Sangoma_cards_no >= MAX_SDLA_CARD){
		DEBUG_EVENT("%s?: Exceed muximum number of Sangoma cards (%d)!\n",
				SDLA_DEV_NAME, Sangoma_cards_no);
		return NULL;
	}
    	if (Sangoma_PCI_cards_no >= MAX_S514_CARDS){
		DEBUG_EVENT("%s?: Exceed muximum number of Sangoma PCI cards (%d)!\n",
				SDLA_DEV_NAME, Sangoma_cards_no);
		return NULL;
    	}	
#if 0
    	if ((device_id & 0xFFFF) != V3_VENDOR_ID || 
	    (device_id >>16) & 0xFFFF != V3_DEVICE_ID){
		return NULL;
	}
#endif
    	PCI_vendor = pci_conf_read(config_id, PCI_VENDOR_ID_WORD);
    	PCI_subsys_vendor = pci_conf_read(config_id, PCI_SUBSYS_VENDOR_WORD);
    	PCI_device = pci_conf_read(config_id, PCI_DEVICE_ID_WORD);
     	/* Check ivendor and subvendor ID */
    	if (!((PCI_vendor == V3_VENDOR_ID && PCI_subsys_vendor == SANGOMA_SUBSYS_VENDOR) ||
    	      (PCI_vendor == PCI_VENDOR_ID_GSI && PCI_device == PCI_DEVICE_ID_GSI_PULSAR))){
		DEBUG_KERNEL("%s: not Sangoma PCI card (PCI_vendor=%04x,PCI_subsys_vendor=%04x)!\n",
				SDLA_DEV_NAME, PCI_vendor, PCI_subsys_vendor);
		return NULL;
    	}

	adaper_type = pci_read_config(dev, PCI_SUBSYS_ID_WORD, 2);
	switch(vendor){
	case V3_VENDOR_ID:
		switch(adapter_tyoe){
		case S5141_ADPTR_1_CPU_SERIAL:
    			return ("Sangoma S5141/FT1 (Single CPU) adapter");
			break;
		case S5142_ADPTR_2_CPU_SERIAL:
    			return ("Sangoma S5142 (Dual CPU) adapter");
			break;
		case S5143_ADPTR_1_CPU_FT1:
    			return ("Sangoma S5143 (Single CPU) FT1 adapter");
			break;
		case S5144_ADPTR_1_CPU_T1E1:
		case S5148_ADPTR_1_CPU_T1E1:
    			return ("Sangoma S5144 (Single CPU) T1/E1 adapter");
			break;
		case S5145_ADPTR_1_CPU_56K:		
    			return ("Sangoma S5145 (Single CPU) 56K adapter");
			break;
		case S5147_ADPTR_2_CPU_T1E1:
    			return ("Sangoma S5147 (Dual CPU) T1/E1 adapter");
			break;
		}
		break;

	case PCI_VENDOR_ID_GSI:
    		return ("Sangoma S518 ADSL adapter");
		break;
	}	

	DEBUG_KERNEL("%s: not Sangoma PCI card (PCI_vendor=%04x,PCI_subsys_vendor=%04x)!\n",
				SDLA_DEV_NAME, PCI_vendor, PCI_subsys_vendor);

    	return NULL;
}

static void sdla_pci_attach(pcici_t config_id, int pci_unit)
{
    	sdladev_t*	pci_cfg = NULL;
    	sdladev_t*	pci_cfg_nxt = NULL;
    	uint16_t 	PCI_vendor, PCI_device;
    	uint16_t 	PCI_subsys_vendor;
    	uint16_t 	adapter_type;
    	int		type = 0, dual_cpu = 0;

    	PCI_vendor = pci_conf_read(config_id, PCI_VENDOR_ID_WORD);
    	PCI_subsys_vendor = pci_conf_read(config_id, PCI_SUBSYS_VENDOR_WORD);
    	PCI_device = pci_conf_read(config_id, PCI_DEVICE_ID_WORD);
	adaper_type = pci_read_config(dev, PCI_SUBSYS_ID_WORD, 2);
    	/* Check ivendor and subvendor ID */
    	if (PCI_vendor == V3_VENDOR_ID && PCI_subsys_vendor == SANGOMA_SUBSYS_VENDOR){
		type = SDLA_S514;
		if (adapter_type == S5148_ADPTR_1_CPU_T1E1){
			adapter_type = S5144_ADPTR_1_CPU_T1E1;
		}
    	}else if (PCI_vendor == PCI_VENDOR_ID_GSI && PCI_device == PCI_DEVICE_ID_GSI_PULSAR){
		type = SDLA_ADSL;
		adapter_type = S518_ADPTR_1_CPU_ADSL;
    	}else{
		return;
    	}

    	if (Sangoma_cards_no+1 >= MAX_SDLA_CARD){
		DEBUG_EVENT("%s%d: Exceed muximum number of Sangoma cards (%d)!\n",
				SDLA_DEV_NAME, pci_unit, Sangoma_cards_no);
		return;
    	}

    	/* Check number of CPUs */
	switch(adapter_type){
	case S5142_ADPTR_2_CPU_SERIAL:
	case S5147_ADPTR_2_CPU_T1E1:
		dual_cpu = 1;
		break;
	}

    	/* Map interrupt handler for first CPU */
    	pci_cfg = &sdladev[Sangoma_cards_no];
    	pci_cfg->type          = type;
    	pci_cfg->adapter_type  = adapter_type;
    	pci_cfg->vendor_id     = PCI_vendor;
    	pci_cfg->device_id     = PCI_device;
    	pci_cfg->isr           = NULL;
    	pci_cfg->irq           = 0;	/* FIXME: Find irq! */
    	pci_cfg->u.pci.pci_dev = config_id;
    	pci_cfg->u.pci.cpu_no  = SDLA_CPU_A;
    	if (!pci_map_int(config_id, sdla_pci_intr, (void*)pci_cfg, &net_imask)){
		DEBUG_EVENT("%s%d: Couldn't map interrupt for cpu A!\n", 
					SDLA_DEV_NAME, pci_unit);
		return;
    	}
 
    	if (dual_cpu){
    		/* Map interrupt handler for second CPU */
    		pci_cfg_nxt = &sdladev[Sangoma_cards_no+1];
    		pci_cfg_nxt->type          = type;
    		pci_cfg_nxt->adapter_type  = adapter_type;
    		pci_cfg_nxt->vendor_id	   = PCI_vendor;
    		pci_cfg_nxt->device_id     = PCI_device;
    		pci_cfg_nxt->isr           = NULL;
    		pci_cfg_nxt->irq           = 0;	/* FIXME: Find irq! */
    		pci_cfg_nxt->u.pci.pci_dev = config_id;
    		pci_cfg_nxt->u.pci.cpu_no  = SDLA_CPU_B;
    		if (!pci_map_int(config_id, sdla_pci_intr, (void*)pci_cfg_nxt, &net_imask)){
			DEBUG_EVENT("%s%d: Couldn't map interrupt for cpu B!\n", 
						SDLA_DEV_NAME, pci_unit);
			return;
    		}
    	}

    	/* Increment number of SDLA adapters, include CPU_A and CPU_B */
    	Sangoma_cards_no += (dual_cpu) ? 2 : 1;
    	/* Incremenet number of S514 cards */
    	Sangoma_PCI_cards_no++;
	/* Increment number of Sangoma devices */
	switch(adapter_type){
	case S5141_ADPTR_1_CPU_SERIAL:
		Sangoma_devices_no += (dual_cpu) ? 2 : 1;
		break;
	case S5142_ADPTR_2_CPU_SERIAL:
		Sangoma_devices_no += (dual_cpu) ? 4 : 2;
		break;
	case S5143_ADPTR_1_CPU_FT1:
	case S5144_ADPTR_1_CPU_T1E1:
	/*case S5148_ADPTR_1_CPU_T1E1:*/
	case S5145_ADPTR_1_CPU_56K:		
	case S518_ADPTR_1_CPU_ADSL:
		Sangoma_devices_no ++;
		break;
	case S5147_ADPTR_2_CPU_T1E1:
		Sangoma_devices_no += 2;
		break;
	}
	return;
}

#endif	/* FreeBSD-3.x */

/*
 ****************************************************************************
 * sdla_pci_intr() - Dummy PCI device interrupt handler.
 */
void sdla_pci_intr(void *arg)
{
    	sdladev_t* 		pci_cfg = (sdladev_t*)arg;

#if (__FreeBSD_version >= 450000)
	DEBUG_EVENT("sdla_pci_intr: should never happened!\n");
#endif

	DEBUG_KERNEL("Interrupt from Sangoma PCI adapter!\n");
	WAN_ASSERT1(pci_cfg == NULL);

    	/* Received interrupt not for Sangoma card */
    	if (pci_cfg->type != SDLA_S514 && pci_cfg->type != SDLA_ADSL){
		DEBUG_EVENT("Interrupt for unknown Sangoma device!\n");
		return;
	}
    	if (pci_cfg->isr != NULL && pci_cfg->arg != NULL){
		pci_cfg->isr(pci_cfg->arg); 
    	}
}

/****** End *****************************************************************/
