

%{

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include "wanhdlcy.h"
#include "wanhdlc.h"

#define YYLMAX		512
#define token(x)	x
#define END(v)		(v-1+sizeof(v)/sizeof(v[0]))
#define YY_NO_UNPUT
#undef YY_INPUT
#define YY_INPUT(buf,result,max_size)			\
	{						\
		char c;				\
		if (targv < arglim){			\
			c = targv[0][offset++];		\
			if (c  != '\0'){		\
				buf[0] = c;		\
			}else{				\
				targv++;		\
				offset = 0;		\
				buf[0] = ' ';		\
			}				\
			result = 1;			\
		}else{					\
			result = 0;			\
		}					\
	}

extern char**	targv;
extern char**	arglim;
static unsigned offset = 0;

extern int yylineno;
char idstr[YYLMAX];

extern unsigned long convert_str(char* str, int type);
extern void yyerror(char*);
void yy_reset(void);

extern char* wanhdlc_version(void);
int wanhdlc_help(void);
static int lookup_name(void);
static void s_lookup(int yylex);

%}

letter			[a-zA-Z_]
exp_letter		[a-zA-Z_\x80-\xFF]
digit			[0-9]
hex			[a-fA-F0-9]
power			[Ee][+-]?{digit}+
float_suffix		(f|F|l|L)
integer_suffix		("u"|"U"|"l"|"L"|"ul"|"lu"|"Ul"|"lU"|"uL"|"Lu"|"UL"|"LU")
letter_or_digit		[a-zA-Z_0-9]
exp_letter_or_digit	[a-zA-Z_0-9\x80-\xFF]
blank			[ \t]
feet_suffix		("ft"|"FT")
db_suffix		("db"|"DB")

%%

{digit}+{feet_suffix}		{ return lookup_name(); }
{digit}+.{digit}+{db_suffix}	{ return lookup_name(); }
{digit}+{db_suffix}		{ return lookup_name(); }

{digit}{digit}*{integer_suffix}?	{ /* decimal */ 
				s_lookup(token(DEC_CONSTANT));
				return token(DEC_CONSTANT);
				}

0[xX]{hex}+{integer_suffix}?  {	/* hex */
				s_lookup(token(HEX_CONSTANT));
				return token(HEX_CONSTANT);
				}

{letter}{letter_or_digit}*	{ return lookup_name(); }

[\n]				{ 
					yylineno++; 
				}
[\t ]				;

"-h"|"--help"			{ wanhdlc_help(); }

"="				{ return ('='); }

.				{ return token(yytext[0]); }


%%

static struct table_arg_t {
	char*	name;
	int 	value;
	char*	descr;
} table_arg[] = {
	{"CHANNEL",		token(HDLC_CHANNEL),		"Description" },
	{"CIR",			token(HDLC_FR_PVC_CIR),		"Description" },
	{"CISCO",		token(HDLC_CISCO),		"Description" },
	{"CREATE",		token(HDLC_FR_PVC_ADD),		"Description" },
	{"DCE",			token(HDLC_FR_DCE),		"Description" },
	{"DELETE",		token(HDLC_FR_PVC_DEL),		"Description" },
	{"EXT",			token(HDLC_CLOCKING_EXT),	"Description" },
	{"FR",			token(HDLC_FR),			"Description" },
	{"FRAME",		token(HDLC_FRAME),		"Description" },
	{"INT",			token(HDLC_CLOCKING_INT),	"Description" },
	{"INVERSE_ARP",		token(HDLC_FR_PVC_INVERSE_ARP),	"Description" },
	{"LBO",			token(HDLC_LBO),		"Description" },
	{"LCODE",		token(HDLC_LCODE),		"Description" },
	{"LMI",			token(HDLC_FR_LMI),		"Description" },
	{"MEDIA",		token(HDLC_MEDIA),		"Description" },
	{"MTU",			token(HDLC_MTU),		"Description" },
	{"N391",		token(HDLC_FR_N391),		"Description" },
	{"N392",		token(HDLC_FR_N392),		"Description" },
	{"N393",		token(HDLC_FR_N393),		"Description" },
	{"PPP",			token(HDLC_PPP),		"Description" },
	{"RS232",		token(HDLC_INTERFACE_RS232),	"Description" },
	{"T391",		token(HDLC_FR_T391),		"Description" },
	{"T392",		token(HDLC_FR_T392),		"Description" },
	{"TECLOCK",		token(HDLC_TECLOCK),		"Description" },
	{"TTL",			token(HDLC_TTL),		"Description" },
	{"UDPPORT",		token(HDLC_UDPPORT),		"Description" },
	{"V35",			token(HDLC_INTERFACE_V35),	"Description" },
};

static struct table_val_t {
	char*	name;
	int 	value;
} table_val[] = {
	{"0db",			token(HDLC_T1_LBO_0_DB) },
	{"110ft",		token(HDLC_T1_LBO_0_110) },
	{"15db",		token(HDLC_T1_LBO_15_DB) },
	{"22.5db",		token(HDLC_T1_LBO_225_DB) },
	{"220ft",		token(HDLC_T1_LBO_110_220) },
	{"330ft",		token(HDLC_T1_LBO_220_330) },
	{"440ft",		token(HDLC_T1_LBO_330_440) },
	{"550ft",		token(HDLC_T1_LBO_440_550) },
	{"660ft",		token(HDLC_T1_LBO_550_660) },
	{"7.5db",		token(HDLC_T1_LBO_75_DB) },
	{"ALL",			token(HDLC_T1E1_CHANNEL_ALL) },
	{"AMI",			token(HDLC_LCODE_AMI) },
	{"ANSI",		token(HDLC_FR_LMI_ANSI) },
	{"B8ZS",		token(HDLC_LCODE_B8ZS) },
	{"CCITT",		token(HDLC_FR_LMI_CCITT) },
	{"CRC4",		token(HDLC_FRAME_CRC4) },
	{"D4",			token(HDLC_FRAME_D4) },
	{"E1",			token(HDLC_MEDIA_E1) },
	{"ESF",			token(HDLC_FRAME_ESF) },
	{"HDB3",		token(HDLC_LCODE_HDB3) },
	{"INTERVAL",		token(HDLC_CISCO_INTERVAL) },
	{"MASTER",		token(HDLC_TECLOCK_MASTER) },
	{"NCRC4",		token(HDLC_FRAME_NCRC4) },
	{"NONE",		token(HDLC_FR_LMI_NONE) },
	{"NORMAL",		token(HDLC_TECLOCK_NORMAL) },
	{"T1",			token(HDLC_MEDIA_T1) },
	{"TIMEOUT",		token(HDLC_CISCO_TIMEOUT) },
	{"UNFRAMED",		token(HDLC_FRAME_UNFRAMED) },
};

static int lookup_name(void)
{
	struct table_arg_t	*first_arg = table_arg,
				*last_arg = END(table_arg),
				*mid_arg;
	struct table_val_t	*first_val = table_val,
				*last_val = END(table_val),
				*mid_val;
	int cmp = 0;
	while(first_arg <= last_arg){
		mid_arg = first_arg + (last_arg - first_arg) / 2;
		if ((cmp = strcasecmp(mid_arg->name, yytext)) == 0){
			return mid_arg->value;
		}else if (cmp < 0){
			first_arg = mid_arg + 1;
		}else{
			last_arg = mid_arg - 1;
		}
	}

	while(first_val <= last_val){
		mid_val = first_val + (last_val - first_val) / 2;
		if ((cmp = strcasecmp(mid_val->name, yytext)) == 0){
			return mid_val->value;
		}else if (cmp < 0){
			first_val = mid_val + 1;
		}else{
			last_val = mid_val - 1;
		}
	}
	s_lookup(token(CHAR_STRING));
	return token(CHAR_STRING);
}


#ifndef YYSTYPE
static char* strsave(char* str)
{
	long	len = strlen(str);
	char*	new_str = NULL;
	
	if (len > YYLMAX){
		yyerror("Line is to long!");
		return NULL;
	}
	new_str = calloc(len, sizeof(char));
	strcpy(new_str, str);
	return new_str;
}
#endif

static void s_lookup(int yylex)
{
#ifdef YYSTYPE
	switch(yylex){
	case DEC_CONSTANT:
	case HEX_CONSTANT:
		memset(idstr, 0, YYLMAX*sizeof(char));
		strcpy(idstr, yytext);
		yylval.val = convert_str(idstr, yylex);
		//yylval.str = idstr;
		break;
	case CHAR_STRING:
		memset(idstr, 0, YYLMAX*sizeof(char));
		strcpy(idstr, yytext);
		yylval.str = idstr;
		break;
	}	
#else
	yylval = strsave(yytext);
#endif
}
 
void yy_reset()
{
	yylineno = 1;
	yy_start = 0;
	yyleng = 0;
}

int yywrap(void)
{
	return 1;
}

int wanhdlc_help(void)
{

	wanhdlc_version();
	printf("\nwanhdlc version %s\n"
"Author:	Alex Feldman	<al.feldman@sangoma.com>\n"
"Copyright (C) 2004 Sangoma Technologies Inc.\n", wanhdlc_version());
	
#if defined(__FreeBSD__) || defined(__OpenBSD__)
	printf(
"Usage:\n"
"    wanhdlc INTERFACE [t1|e1]  [LCODE] [FRAME] [LBO] [TECLOCK] [channel=LIST]\n"
"    wanhdlc INTERFACE [v35|rs232] [int=<value> |ext]\n"
"    wanhdlc INTERFACE [mtu=<value>] [ttl=<value>] [udpport=<value>]\n"
"    wanhdlc INTERFACE [PROTOCOL]\n"
"\n"
"LCODE := b8zs|hdb3|ami\n"
"FRAME := esf|d4|crc4|ncrc4\n"
"LBO := 0db| 7.5db| 15db| 22.5db| 110ft| 220ft| 330ft| 440ft| 550ft| 660ft\n"
"TECLOCK := master|normal\n"
"LIST := a.b-c|ALL\n"
"PROTOCOL := cisco | ppp\n");

#elif defined(__LINUX__)
# if defined(KERNEL_V26)
	printf(
"Usage:\n"
"    wanhdlc INTERFACE - Display current full configuration for this interface\n"
"    wanhdlc INTERFACE [t1|e1] [LCODE] [FRAME] [LBO] [TECLOCK] [channel=LIST]\n"
"    wanhdlc INTERFACE [v35|rs232] [int=<value> |ext]\n"
"    wanhdlc INTERFACE [mtu=<value>] [ttl=<value>] [udpport=<value>]\n"
"    wanhdlc INTERFACE [PROTOCOL]\n"
"    sethdlc INTERFACE create DLCI\n"
"    sethdlc INTERFACE delete DLCI\n"
"\n"
"LCODE := b8zs|hdb3|ami\n"
"FRAME := esf|d4|crc4|ncrc4|unframed\n"
"LBO := 0db| 7.5db| 15db| 22.5db| 110ft| 220ft| 330ft| 440ft| 550ft| 660ft\n"
"TECLOCK := master|normal\n"
"LIST := a.b-c|ALL\n"
"PROTOCOL := cisco [interval val] [timeout val] |\n"
"            fr [lmi LMI] |\n"
"            ppp | x25\n"
"ENCODING := nrz | nrzi | fm-mark | fm-space | manchester\n"
"PARITY := no-parity| crc16| crc16-pr0| crc16-itu| crc16-itu-pr0| crc32-itu\n"
"LMI := none | ansi [LMI_SPEC] | ccitt [LMI_SPEC]\n"
"LMI_SPEC := [dce] [t391 val] [t392 val] [n391 val] [n392 val] [n393 val]\n\n");

# else
	printf(
"Usage:\n"
"    wanhdlc INTERFACE - Display current full configuration for this interface\n"
"    wanhdlc INTERFACE [t1|e1] [LCODE] [FRAME] [LBO] [TECLOCK] [channel=LIST]\n"
"    wanhdlc INTERFACE [v35|rs232] [int=<value> |ext]\n"
"    wanhdlc INTERFACE [mtu=<value>] [ttl=<value>] [udpport=<value>]\n"
"    wanhdlc INTERFACE [PROTOCOL]\n"
"    sethdlc INTERFACE PVC create DLCI [inverse-arp] [CIR]\n"
"    sethdlc INTERFACE PVC delete DLCI\n"
"\n"
"LCODE := b8zs|hdb3|ami\n"
"FRAME := esf|d4|crc4|ncrc4|unframed\n"
"LBO := 0db| 7.5db| 15db| 22.5db| 110ft| 220ft| 330ft| 440ft| 550ft| 660ft\n"
"TECLOCK := master|normal\n"
"LIST := a.b-c|ALL\n"
"PROTOCOL := cisco [interval val] [timeout val] |\n"
"            fr [lmi LMI] |\n"
"            ppp | x25\n"
"ENCODING := nrz | nrzi | fm-mark | fm-space | manchester\n"
"PARITY := no-parity| crc16| crc16-pr0| crc16-itu| crc16-itu-pr0| crc32-itu\n"
"LMI := none | ansi [LMI_SPEC] | ccitt [LMI_SPEC]\n"
"LMI_SPEC := [dce] [t391 val] [t392 val] [n391 val] [n392 val] [n393 val]\n"
"PVC := pvc0 | pvc1 | ... | pvc9999\n"
"CIR := cir BC INTERVAL MARK\n\n");
# endif
#endif

#if 0
	fprintf(stderr, "\nsethdlc version 2.0C"
"Copyright (C) 2000 - 2001 Krzysztof Halasa <khc@pm.waw.pl>"
"\n"
"Usage:\n"
"    sethdlc INTERFACE [PHYSICAL] [clock CLOCK] [LOOPBACK] [slotmap SLOTMAP]\n"
"    sethdlc INTERFACE [PROTOCOL]\n"
"    sethdlc INTERFACE PVC create DLCI [inverse-arp] [CIR]\n"
"    sethdlc INTERFACE PVC delete DLCI\n"
"    sethdlc INTERFACE private\n"
"\n"
"PHYSICAL := v24 | v35 | x21 | e1 | t1"
"CLOCK := int [rate RATE] | ext | txint [rate RATE] | txfromrx"
"LOOPBACK := loopback | lb"
"\n"
"PROTOCOL := hdlc [ENCODING] [PARITY] |"
"            cisco [interval val] [timeout val] |"
"            fr [lmi LMI] |"
"            ppp |"
"            x25"
"\n"
"ENCODING := nrz | nrzi | fm-mark | fm-space | manchester"
"PARITY := no-parity | crc16 | crc16-pr0 | crc16-itu | crc16-itu-pr0 | crc32-itu"
"LMI := none | ansi [LMI_SPEC] | ccitt [LMI_SPEC]"
"LMI_SPEC := [dce] [t391 val] [t392 val] [n391 val] [n392 val] [n393 val]"
"\n"
"PVC := pvc0 | pvc1 | ... | pvc9999"
"CIR := cir BC INTERVAL MARK"
);
#endif
	exit(0);
}

