


%{

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <string.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#if defined(__FreeBSD__) || defined(__OpenBSD__)
# include <netinet/ip.h>
# include <netinet/udp.h>  
# include <net/if.h>
# include <net/wanpipe.h>
# include <net/wanpipe.h>
#elif defined(__LINUX__)
# include <linux/if.h>
# include <linux/hdlc.h>
# include <linux/if_wanpipe.h>
# include <linux/wanpipe.h>
#endif
#include "wanhdlc.h"

extern char**	targv;
int 		yylineno = 1;
extern char 	yytext[];

extern int yylex(void);
unsigned long convert_str(char* str, int type);
static unsigned long timeslot_mask, first_ts;

extern int wanhdlc_ifname(char* ifname);
extern int wanhdlc_channel_mask(unsigned long first_ts, unsigned long last_ts);
extern int wanhdlc_te_config(int arg);
extern int wanhdlc_set_te_config(int arg, unsigned long value);
extern int wanhdlc_dev_config(int arg, int value);
extern int wanhdlc_proto_cisco(int arg);
extern int wanhdlc_set_proto_cisco(int arg, unsigned long value);
extern int wanhdlc_proto_fr(int arg);
extern int wanhdlc_set_proto_fr(int arg, unsigned long value);
extern int wanhdlc_fr_pvc(int arg);
extern int wanhdlc_set_fr_pvc(int arg, unsigned long value, char*);
extern int wanhdlc_proto_ppp(int arg);

#if 0
t1e1_configuration	: HDLC_MEDIA '=' media_value
			| HDLC_LCODE '=' lcode_value
			| HDLC_FRAME '=' frame_value
			| HDLC_CHANNEL '=' channels
			| HDLC_TECLOCK '=' teclock_value
			| HDLC_LBO '=' lbo_value
			;
#endif
%}

%union {
#define YYSTYPE YYSTYPE
	char*		str;
	unsigned long	val;
}


%token CHAR_STRING
%token DEC_CONSTANT
%token HEX_CONSTANT

%token HDLC_T1E1
%token HDLC_MEDIA
%token HDLC_LCODE
%token HDLC_FRAME
%token HDLC_LBO
%token HDLC_TECLOCK
%token HDLC_CHANNEL
%token HDLC_INTERFACE
%token HDLC_CLOCKING
%token HDLC_MTU
%token HDLC_UDPPORT
%token HDLC_TTL
%token HDLC_CISCO
%token HDLC_FR
%token HDLC_FR_LMI
%token HDLC_PPP

%token HDLC_FR_PVC
%token HDLC_FR_PVC_NAME
%token HDLC_FR_PVC_ADD
%token HDLC_FR_PVC_DEL
%token HDLC_FR_PVC_DLCI
%token HDLC_FR_PVC_INVERSE_ARP
%token HDLC_FR_PVC_CIR
%token HDLC_FR_PVC_BC
%token HDLC_FR_PVC_INTERVAL
%token HDLC_FR_PVC_NFMARK

%token HDLC_MEDIA_T1
%token HDLC_MEDIA_E1

%token HDLC_LCODE_B8ZS
%token HDLC_LCODE_HDB3
%token HDLC_LCODE_AMI
%token HDLC_FRAME_ESF
%token HDLC_FRAME_D4
%token HDLC_FRAME_CRC4
%token HDLC_FRAME_NCRC4
%token HDLC_FRAME_UNFRAMED
%token HDLC_T1_LBO_0_DB
%token HDLC_T1_LBO_75_DB
%token HDLC_T1_LBO_15_DB
%token HDLC_T1_LBO_225_DB
%token HDLC_T1_LBO_0_110
%token HDLC_T1_LBO_110_220
%token HDLC_T1_LBO_220_330
%token HDLC_T1_LBO_330_440
%token HDLC_T1_LBO_440_550
%token HDLC_T1_LBO_550_660
%token HDLC_T1E1_CHANNEL_ALL
%token HDLC_TECLOCK_MASTER
%token HDLC_TECLOCK_NORMAL
%token HDLC_INTERFACE_V35
%token HDLC_INTERFACE_RS232
%token HDLC_CLOCKING_INT
%token HDLC_CLOCKING_EXT
%token HDLC_CISCO_INTERVAL
%token HDLC_CISCO_TIMEOUT
%token HDLC_FR_LMI_NONE
%token HDLC_FR_LMI_ANSI
%token HDLC_FR_LMI_CCITT
%token HDLC_FR_T391
%token HDLC_FR_T392
%token HDLC_FR_N391
%token HDLC_FR_N392
%token HDLC_FR_N393
%token HDLC_FR_DCE
%token '='

/* hardware arguments */


%%

hdlc_configuration	: interface_name
			| interface_name interface_configuration
			;

interface_name		: CHAR_STRING
			  { wanhdlc_ifname($<str>1); }
			;

interface_configuration	: t1e1_configurations
			| device_configurations
			| cisco_configuration
			| fr_configuration
			| fr_pvc_configuration
			| ppp_configuration
			;

device_configurations	: device_configuration
			| device_configuration device_configurations
			;

device_configuration	: HDLC_INTERFACE_V35
			  { wanhdlc_dev_config(HDLC_INTERFACE, WANOPT_V35); }
			| HDLC_INTERFACE_RS232
			  { wanhdlc_dev_config(HDLC_INTERFACE, WANOPT_RS232); }
			| HDLC_CLOCKING_INT '=' DEC_CONSTANT
			  { wanhdlc_dev_config(HDLC_CLOCKING, $<val>3); }
			| HDLC_CLOCKING_EXT
			  { wanhdlc_dev_config(HDLC_CLOCKING, 0); }
			| HDLC_MTU '=' DEC_CONSTANT
			  { wanhdlc_dev_config(HDLC_MTU, $<val>3); }
			| HDLC_UDPPORT '=' DEC_CONSTANT
			  { wanhdlc_dev_config(HDLC_UDPPORT, $<val>3); }
			| HDLC_TTL '=' DEC_CONSTANT
			  { wanhdlc_dev_config(HDLC_TTL, $<val>3); }
			;

t1e1_configurations	: t1e1_configuration
			  { wanhdlc_te_config(HDLC_T1E1); }
			| t1e1_configuration t1e1_configurations
			;

t1e1_configuration	: media_value
			| lcode_value
			| frame_value
			| teclock_value
			| HDLC_CHANNEL '=' channels
			| lbo_value
			;


media_value		: HDLC_MEDIA_T1
			  { wanhdlc_set_te_config(HDLC_MEDIA, WANOPT_MEDIA_T1); }
			| HDLC_MEDIA_E1
			  { wanhdlc_set_te_config(HDLC_MEDIA, WANOPT_MEDIA_E1); }
			;

lcode_value		: HDLC_LCODE_B8ZS
			  { wanhdlc_set_te_config(HDLC_LCODE, WANOPT_LC_B8ZS); }
			| HDLC_LCODE_HDB3
			  { wanhdlc_set_te_config(HDLC_LCODE, WANOPT_LC_HDB3); }
			| HDLC_LCODE_AMI
			  { wanhdlc_set_te_config(HDLC_LCODE, WANOPT_LC_AMI); }
			;
	
frame_value		: HDLC_FRAME_ESF
			  { wanhdlc_set_te_config(HDLC_FRAME, WANOPT_FR_ESF); }
			| HDLC_FRAME_D4
			  { wanhdlc_set_te_config(HDLC_FRAME, WANOPT_FR_D4); }
			| HDLC_FRAME_CRC4
			  { wanhdlc_set_te_config(HDLC_FRAME, WANOPT_FR_CRC4); }
			| HDLC_FRAME_NCRC4
			  { wanhdlc_set_te_config(HDLC_FRAME, WANOPT_FR_NCRC4); }
			| HDLC_FRAME_UNFRAMED
			  { wanhdlc_set_te_config(HDLC_FRAME, WANOPT_FR_UNFRAMED); }
			;

teclock_value		: HDLC_TECLOCK_NORMAL
			  { wanhdlc_set_te_config(HDLC_TECLOCK, WANOPT_NORMAL_CLK); }
			| HDLC_TECLOCK_MASTER
			  { wanhdlc_set_te_config(HDLC_TECLOCK, WANOPT_MASTER_CLK); }
			;

lbo_value		: HDLC_T1_LBO_0_DB
			  { wanhdlc_set_te_config(HDLC_LBO, WANOPT_T1_LBO_0_DB); }
			| HDLC_T1_LBO_75_DB
			  { wanhdlc_set_te_config(HDLC_LBO, WANOPT_T1_LBO_75_DB); }
			| HDLC_T1_LBO_15_DB
			  { wanhdlc_set_te_config(HDLC_LBO, WANOPT_T1_LBO_15_DB); }
			| HDLC_T1_LBO_225_DB
			  { wanhdlc_set_te_config(HDLC_LBO, WANOPT_T1_LBO_225_DB); }
			| HDLC_T1_LBO_0_110
			  { wanhdlc_set_te_config(HDLC_LBO, WANOPT_T1_0_110); }
			| HDLC_T1_LBO_110_220
			  { wanhdlc_set_te_config(HDLC_LBO, WANOPT_T1_110_220); }
			| HDLC_T1_LBO_220_330
			  { wanhdlc_set_te_config(HDLC_LBO, WANOPT_T1_220_330); }
			| HDLC_T1_LBO_330_440
			  { wanhdlc_set_te_config(HDLC_LBO, WANOPT_T1_330_440); }
			| HDLC_T1_LBO_440_550
			  { wanhdlc_set_te_config(HDLC_LBO, WANOPT_T1_440_550); }
			| HDLC_T1_LBO_550_660
			  { wanhdlc_set_te_config(HDLC_LBO, WANOPT_T1_550_660); }
			;

channels		: HDLC_T1E1_CHANNEL_ALL
			  { wanhdlc_set_te_config(HDLC_CHANNEL, ENABLE_ALL_CHANNELS); }
			| channel_list
			  { wanhdlc_set_te_config(HDLC_CHANNEL, timeslot_mask); }
			;

channel_list		: channel_value
			| channel_list '.' channel_value
			;

channel_value		: DEC_CONSTANT
			  { timeslot_mask |= wanhdlc_channel_mask($<val>1, $<val>1); } 
			| DEC_CONSTANT '-'
			  { first_ts = $<val>1; }
				DEC_CONSTANT
			  		{ timeslot_mask |= wanhdlc_channel_mask(first_ts, $<val>4); }
			;

cisco_configuration	: HDLC_CISCO cisco_params
			  { wanhdlc_proto_cisco(HDLC_CISCO); }
			;

cisco_params		: 
			| HDLC_CISCO_INTERVAL DEC_CONSTANT cisco_params
			  { wanhdlc_set_proto_cisco(HDLC_CISCO_INTERVAL, $<val>2); }
			| HDLC_CISCO_TIMEOUT DEC_CONSTANT cisco_params
			  { wanhdlc_set_proto_cisco(HDLC_CISCO_TIMEOUT, $<val>2); }
			;

fr_configuration	: HDLC_FR fr_params
			  { wanhdlc_proto_fr(HDLC_FR); }
			;

fr_params		:
			| HDLC_FR_LMI HDLC_FR_LMI_NONE 
			  { wanhdlc_set_proto_fr(HDLC_FR_LMI, WAN_LMI_NONE); }
			| HDLC_FR_LMI fr_lmi_type 
			  { wanhdlc_set_proto_fr(HDLC_FR_LMI, $<val>2); }
					fr_lmi_spec
			;

fr_lmi_type		: HDLC_FR_LMI_ANSI
			  { $<val>$ = WAN_LMI_ANSI; }
			| HDLC_FR_LMI_CCITT
			  { $<val>$ = WAN_LMI_CCITT; }
			;

fr_lmi_spec		: 
			| HDLC_FR_T391 DEC_CONSTANT fr_lmi_spec
			  { wanhdlc_set_proto_fr(HDLC_FR_T391, $<val>2); }
			| HDLC_FR_T392 DEC_CONSTANT fr_lmi_spec
			  { wanhdlc_set_proto_fr(HDLC_FR_T392, $<val>2); }
			| HDLC_FR_N391 DEC_CONSTANT fr_lmi_spec
			  { wanhdlc_set_proto_fr(HDLC_FR_N391, $<val>2); }
			| HDLC_FR_N392 DEC_CONSTANT fr_lmi_spec
			  { wanhdlc_set_proto_fr(HDLC_FR_N392, $<val>2); }
			| HDLC_FR_N393 DEC_CONSTANT fr_lmi_spec
			  { wanhdlc_set_proto_fr(HDLC_FR_N393, $<val>2); }
			| HDLC_FR_DCE fr_lmi_spec
			  { wanhdlc_set_proto_fr(HDLC_FR_DCE, 0); }
			;

fr_pvc_configuration	: CHAR_STRING fr_pvc_action
			  { wanhdlc_set_fr_pvc(HDLC_FR_PVC_NAME, 0, $<str>1); }
					DEC_CONSTANT 
			  { wanhdlc_set_fr_pvc(HDLC_FR_PVC_DLCI, $<val>4, NULL); }
						fr_pvc_params
			  { wanhdlc_fr_pvc(HDLC_FR_PVC); }

			;

fr_pvc_action		: HDLC_FR_PVC_ADD
			  { wanhdlc_set_fr_pvc(HDLC_FR_PVC_ADD, 0, NULL); }
			| HDLC_FR_PVC_DEL
			  { wanhdlc_set_fr_pvc(HDLC_FR_PVC_DEL, 0, NULL); }
			;

fr_pvc_params		: HDLC_FR_PVC_INVERSE_ARP fr_pvc_cir
			  { wanhdlc_set_fr_pvc(HDLC_FR_PVC_INVERSE_ARP, 0, NULL); }
			| fr_pvc_cir
			;

fr_pvc_cir		:
			| HDLC_FR_PVC_CIR DEC_CONSTANT
			  { wanhdlc_set_fr_pvc(HDLC_FR_PVC_BC, $<val>2, NULL); }
				DEC_CONSTANT
			  { wanhdlc_set_fr_pvc(HDLC_FR_PVC_INTERVAL, $<val>4, NULL); }
					DEC_CONSTANT
			  { wanhdlc_set_fr_pvc(HDLC_FR_PVC_NFMARK, $<val>6, NULL); }
			;


ppp_configuration	: HDLC_PPP
			  { wanhdlc_proto_ppp(HDLC_PPP); }
			;



%%

	
unsigned long convert_str(char* str, int type)
{
	unsigned long value = 0;
	switch(type){
	case DEC_CONSTANT:
		sscanf(str, "%lu", &value);
		break;
	case HEX_CONSTANT:
		sscanf(str, "%lx", &value);
		break;
	}
	return value;
}

void yyerror(char* msg)
{
	printf("ERROR (%s): Invalid parameter -> %s!\n", msg, (targv-1)[0]);
//	wanhdlc_error(ERR_PARSE, msg, NULL, 0);
}
