/****************************************************************************
* wanpipe_main.c	WANPIPE(tm) OpenSource WANPIPe Driver.
*
* Author:		Alex Feldman	<al.feldman@sangoma.com>
*		
*
* Copyright:		(c) 2004 Sangoma Technologies Inc.
*
*		This program is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
* ============================================================================
* Jan 28, 2004	Alex Feldman	Initial version.
*****************************************************************************/

#if defined (__FreeBSD__) || defined(__OpenBSD__)

# include <net/wanpipe_config.h>
# include <net/wanpipe_version.h>
# include <net/wanpipe_includes.h>
# include <net/sdlapci.h>
# include <net/wanpipe.h>	/* WANPIPE common user API definitions */
# include <net/if_wanpipe_common.h>
# include <net/wanpipe_generic.h>

#else

# include <linux/wanpipe_includes.h>
# include <linux/wanpipe_defines.h>
# include <linux/wanrouter.h>	/* WAN router definitions */
# include <linux/sdladrv.h>
# include <linux/wanpipe.h>	/* WANPIPE common user API definitions */
# include <linux/sdlapci.h>
# include <linux/if_wanpipe.h>
# include <linux/if_wanpipe_common.h>
# include <linux/wanpipe_generic.h>
#endif
#include "cdual514sfm.h"

/***********FOR DEBUGGING PURPOSES*********************************************
static void * dbg_kmalloc(unsigned int size, int prio, int line) {
	int i = 0;
	void * v = kmalloc(size+sizeof(unsigned int)+2*KMEM_SAFETYZONE*8,prio);
	char * c1 = v;	
	c1 += sizeof(unsigned int);
	*((unsigned int *)v) = size;

	for (i = 0; i < KMEM_SAFETYZONE; i++) {
		c1[0] = 'D'; c1[1] = 'E'; c1[2] = 'A'; c1[3] = 'D';
		c1[4] = 'B'; c1[5] = 'E'; c1[6] = 'E'; c1[7] = 'F';
		c1 += 8;
	}
	c1 += size;
	for (i = 0; i < KMEM_SAFETYZONE; i++) {
		c1[0] = 'M'; c1[1] = 'U'; c1[2] = 'N'; c1[3] = 'G';
		c1[4] = 'W'; c1[5] = 'A'; c1[6] = 'L'; c1[7] = 'L';
		c1 += 8;
	}
	v = ((char *)v) + sizeof(unsigned int) + KMEM_SAFETYZONE*8;
	printk(KERN_INFO "line %d  kmalloc(%d,%d) = %p\n",line,size,prio,v);
	return v;
}
static void dbg_kfree(void * v, int line) {
	unsigned int * sp = (unsigned int *)(((char *)v) - (sizeof(unsigned int) + KMEM_SAFETYZONE*8));
	unsigned int size = *sp;
	char * c1 = ((char *)v) - KMEM_SAFETYZONE*8;
	int i = 0;
	for (i = 0; i < KMEM_SAFETYZONE; i++) {
		if (   c1[0] != 'D' || c1[1] != 'E' || c1[2] != 'A' || c1[3] != 'D'
		    || c1[4] != 'B' || c1[5] != 'E' || c1[6] != 'E' || c1[7] != 'F') {
			printk(KERN_INFO "kmalloced block at %p has been corrupted (underrun)!\n",v);
			printk(KERN_INFO " %4x: %2x %2x %2x %2x %2x %2x %2x %2x\n", i*8,
			                c1[0],c1[1],c1[2],c1[3],c1[4],c1[5],c1[6],c1[7] );
		}
		c1 += 8;
	}
	c1 += size;
	for (i = 0; i < KMEM_SAFETYZONE; i++) {
		if (   c1[0] != 'M' || c1[1] != 'U' || c1[2] != 'N' || c1[3] != 'G'
		    || c1[4] != 'W' || c1[5] != 'A' || c1[6] != 'L' || c1[7] != 'L'
		   ) {
			printk(KERN_INFO "kmalloced block at %p has been corrupted (overrun):\n",v);
			printk(KERN_INFO " %4x: %2x %2x %2x %2x %2x %2x %2x %2x\n", i*8,
			                c1[0],c1[1],c1[2],c1[3],c1[4],c1[5],c1[6],c1[7] );
		}
		c1 += 8;
	}
	printk(KERN_INFO "line %d  kfree(%p)\n",line,v);
	v = ((char *)v) - (sizeof(unsigned int) + KMEM_SAFETYZONE*8);
	kfree(v);
}

#define kmalloc(x,y) dbg_kmalloc(x,y,__LINE__)
#define kfree(x) dbg_kfree(x,__LINE__)
******************************************************************************/



/****** Defines & Macros ****************************************************/

#ifdef	_DEBUG_
#define	STATIC
#else
#define	STATIC		static
#endif

#ifndef	CONFIG_WANPIPE_CARDS		/* configurable option */
#define	CONFIG_WANPIPE_CARDS 1
#endif

#define INVALID_ADAPTER_CHECK(card,cfg) \
	{									\
	  u16	adapter_type = 0x00;						\
	  card->hw_iface.getcfg(card->hw, SDLA_ADAPTERTYPE, &adapter_type);	\
	  if (adapter_type == S5144_ADPTR_1_CPU_T1E1 || 			\
	      IS_TE1(cfg)){ 							\
	      	DEBUG_EVENT("%s: Error: Protocol not supported on S514-4 T1/E1 Adapter\n", \
			    	card->devname); 				\
		err = -EPFNOSUPPORT; 						\
		break; 								\
	  }else if (adapter_type == S5145_ADPTR_1_CPU_56K || 			\
	            IS_56K(cfg)){ 						\
	        DEBUG_EVENT("%s: Error: Protocol not supported on S514-5 56K Adapter\n", \
			    	card->devname); 				\
		err = -EPFNOSUPPORT; 						\
		break; 								\
          } \
       }

/****** Function Prototypes *************************************************/

 
/* Module entry points */
int init_module (void);
void cleanup_module (void);

/* WAN link driver entry points */
static int	wanpipe_setup(wan_device_t* wandev);
static int	shutdown(wan_device_t* wandev);
static int	wanpipe_find_same_card(sdla_t* card);
static int 	check_conflicts(sdla_t* card);

/* Miscellaneous functions */
static int wanpipe_ioctl(netdevice_t* dev, struct ifreq* ifr, int cmd);
#if defined(__FreeBSD__) || defined(__OpenBSD__)
STATIC irqreturn_t sdla_isr (void* dev_id);
#elif defined(__LINUX__)
STATIC irqreturn_t sdla_isr (int irq, void* dev_id, struct pt_regs *regs);
extern void disable_irq(unsigned int);
extern void enable_irq(unsigned int);
#endif

static void release_hw  (sdla_t *card);

static int wanpipe_init(void);
static int wanpipe_exit(void);
extern int sdladrv_init(void);
extern int sdladrv_exit(void);

static int wanpipe_get_default_cfg(sdla_t*, wandev_conf_t**, wanif_conf_t**);
extern int wp_mprot_default_devcfg(void*, void*);
extern int wp_mprot_default_ifcfg(void*, void*);
#if defined(__LINUX__)
//extern int wp_xilinx_default_devcfg(void*, void*);
//extern int wp_xilinx_default_ifcfg(void*, void*);
extern int wp_xilinx_default_devcfg(sdla_t* card, wandev_conf_t*);
extern int wp_xilinx_default_ifcfg(sdla_t*, wanif_conf_t*);
#endif
extern netdevice_t* wanpipe_generic_register(sdla_t* card);
extern void wanpipe_generic_unregister(netdevice_t* dev);

static int wanpipe_ioctl_dump(sdla_t* card, void* u_dump);
static int wanpipe_ioctl_hwprobe(netdevice_t* dev, void* u_def);

void debug_print_udp_pkt(unsigned char *data,int len,char trc_enabled, char direction);
/****** Global Data **********************************************************
 * Note: All data must be explicitly initialized!!!
 */

/* private data */
static char drvname[]	= "wanpipe";
static char fullname[]	= "WANPIPE(tm) WAN Generic Driver Module";
static char wanpipe_lite_copyright[]	= "(c) 2004 Sangoma Technologies Inc.";
static int ncards; 

static sdla_t* card_list;	/* adapter data space */
sdla_t* wanpipe_debug = NULL;


	
/******* Kernel Loadable Module Entry Points ********************************/

#if defined(__OpenBSD__)
static struct cdevsw wanpipe_lite_devsw = {
	NULL, NULL, NULL,NULL,NULL,NULL,NULL,NULL,NULL,0,NULL
};
#else
void*	wanpipe_lite_devsw = NULL;
#endif

WAN_MODULE_DEFINE(wanpipe_lite, 
		"Alex Feldman <al.feldman@sangoma.com>",
		"Sangoma WANPIPE: WAN LITE Driver",
		"GPL",
		wanpipe_init,
		wanpipe_exit,
		&wanpipe_lite_devsw);
#if 0
MODULE_AUTHOR ("Alex Feldman <al.feldman@sangoma.com>");
MODULE_DESCRIPTION ("Sangoma WANPIPE: WAN Generic Driver");
MODULE_LICENSE("GPL");

/*============================================================================
 * Module 'insert' entry point.
 * o print announcement
 * o allocate adapter data space
 * o initialize static data
 * o register all cards with WAN router
 * o calibrate SDLA shared memory access delay.
 *
 * Return:	0	Ok
 *		< 0	error.
 * Context:	process
 */
int __init wanpipe_mod_init(void)
{
	return wanpipe_init();
}

/*============================================================================
 * Module 'remove' entry point.
 * o unregister all adapters from the WAN router
 * o release all remaining system resources
 */
void __exit wanpipe_mod_exit(void)
{
	wanpipe_exit();
	return;
}
module_init(wanpipe_mod_init);
module_exit(wanpipe_mod_exit);
#endif

static int wanpipe_init(void)
{
	int i, err = 0;
	sdla_t *card;

	ncards=0;
#if defined(CONFIG_PRODUCT_WANPIPE_GENERIC)
	sdladrv_init();
#endif
	DEBUG_EVENT("%s v%s %s\n",
		fullname, WANPIPE_VERSION, wanpipe_lite_copyright);

	/* Probe for wanpipe cards and return the number found */
	DEBUG_EVENT("wanpipe: Probing for WANPIPE hardware.\n");
	ncards = sdla_hw_probe();
	if (ncards){
		DEBUG_EVENT("wanpipe: Allocating maximum %i devices: wanpipe%i - wanpipe%i.\n",
					ncards,1,ncards);
	}else{
		DEBUG_EVENT("wanpipe: No S514/S508 cards found, unloading modules!\n");
		return -ENODEV;
	}
	card_list=NULL;
	for (i=0;i<ncards;i++){
		wan_device_t* wandev;
		card=wan_malloc(sizeof(sdla_t));
		if (!card){
			DEBUG_EVENT("%s%d: Failed allocate new card!\n", 
					drvname, i+1);
			goto wanpipe_init_done;
		}
		memset(card,0,sizeof(sdla_t));
		sprintf(card->devname, "%s%d", drvname, i+1);
		wandev 			= &card->wandev;
		wandev->magic   	= ROUTER_MAGIC;
		wandev->name    	= card->devname;
		wandev->private 	= card;
		wandev->enable_tx_int	= 0;
		wandev->setup		= NULL;
		wandev->shutdown	= NULL;
		wandev->ioctl   	= NULL;
		wandev->debugging	= NULL;
		card->list	= card_list;
		card_list	= card;
		if ((err = wanpipe_setup(wandev))){
			DEBUG_EVENT("%s: Failed setup new device!\n", 
					card->devname);
			goto wanpipe_init_done;
		}
		wandev->wanpipe_ioctl  	= wanpipe_ioctl;
	}
wanpipe_init_done:
	if (err) wanpipe_exit();
	return err;
}

static int wanpipe_exit(void)
{
	sdla_t *card, *tmp;
	int 	err = 0;

	for (card=card_list;card;){
		if (card->wandev.dev){
			netdevice_t*	dev = card->wandev.dev;
			DEBUG_EVENT("%s: Unregister WANPIPE Generic interface %s\n", 
					card->devname,
					wan_netif_name(dev));
			if (card->wandev.del_if){
				DEBUG_EVENT("%s: Delete interface %s\n", 
						card->devname,
						wan_netif_name(dev));
				card->wandev.del_if(&card->wandev, dev); 
			}
			card->wandev.dev = NULL;
			wanpipe_generic_unregister(dev);
		}
		DEBUG_EVENT("%s: Shutdown device\n", card->devname);
		shutdown(&card->wandev);
		if (card->hw){
			sdla_unregister(&card->hw, card->devname);
		}
		tmp=card->list;
		wan_free(card);
		card=tmp;
	}

	card_list=NULL;
	DEBUG_EVENT("\n");
	DEBUG_EVENT("wanpipe: WANPIPE Generic Modules Unloaded.\n");

#if defined(CONFIG_PRODUCT_WANPIPE_GENERIC)
	err = sdladrv_exit();
#endif

#if defined(WAN_DEBUG_MEM)
	DEBUG_EVENT("wanpipe: Total Mem %d\n",atomic_read(&wan_debug_mem));
#endif	
	return err;
}


/******* WAN Device Driver Entry Points *************************************/

/*============================================================================
 * Setup/configure WAN link driver.
 * o check adapter state
 * o make sure firmware is present in configuration
 * o make sure I/O port and IRQ are specified
 * o make sure I/O region is available
 * o allocate interrupt vector
 * o setup SDLA hardware
 * o call appropriate routine to perform protocol-specific initialization
 * o mark I/O region as used
 * o if this is the first active card, then schedule background task
 *
 * This function is called when router handles ROUTER_SETUP IOCTL. The
 * configuration structure is in kernel memory (including extended data, if
 * any).
 */
 
static int wanpipe_setup (wan_device_t* wandev)
{
	sdla_t*		card = wandev->private;
	wandev_conf_t*	devconf;
	wanif_conf_t*	ifconf;
	int		err = 0;
	u16		irq;
	netdevice_t*	dev;

	card->wandev.piggyback = 0;

	DEBUG_EVENT("\n");
	DEBUG_EVENT("Processing WAN device %s...\n", wandev->name);

	/* Initialize the counters for each wandev 
	 * Used for counting number of times new_if and 
       	 * del_if get called.
	 */
	wandev->del_if_cnt = 0;
	wandev->new_if_cnt = 0;
	wandev->card_type  = card->type;
	card->hw = sdla_register(&card->hw_iface, NULL, card->devname);
	if (card->hw == NULL){
		return -EINVAL;
	}
	/* Check for resource conflicts and setup the
	 * card for piggibacking if necessary */
	card->hw_iface.getcfg(card->hw, SDLA_CARDTYPE, &card->type);
	if (wanpipe_get_default_cfg(card, &devconf, &ifconf)){
		return -EINVAL;
	}
	/* Check for resource conflicts and setup the
	 * card for piggibacking if necessary */
	card->wandev.piggyback = 0;
	switch (card->type){
	case SDLA_S508:
	case SDLA_S514:
		if ((err=check_conflicts(card)) != 0){
			return err;
		}
		break;
	
	case SDLA_ADSL:
	case SDLA_AFT:
		break;
	}

	wandev->config_id  = devconf->config_id;
	
	/* Link the same cards together. I need this only if customer change 
	 * T1 -> E1, in this case I need to change other cpu too. */
	wanpipe_find_same_card(card);

	/* If the current card has already been configured
       	 * or its a piggyback card, do not try to allocate
         * resources.
	 */
	if (!card->wandev.piggyback && !card->configured){

		err = card->hw_iface.setup(card->hw, NULL);
		if (err){
			DEBUG_EVENT("%s: Hardware setup Failed %i\n",
				card->devname,err);
			sdla_unregister(&card->hw, card->devname);
			goto wanpipe_setup_done;
		}
		if (card->type == SDLA_S514 && card->hw_iface.load){
			err = card->hw_iface.load(card->hw,
						 chdlc_firmware,
						 chdlc_firmware_len);
			if (err){
				DEBUG_EVENT("%s: Failed to load firmware (%d)\n",
							card->devname,err);
				card->hw_iface.down(card->hw);
				sdla_unregister(&card->hw, card->devname);
				goto wanpipe_setup_done;
			}
		}

		card->hw_iface.getcfg(card->hw, SDLA_IRQ, &irq);

               	/* Initialize the Spin lock */
#if defined(CONFIG_SMP) || defined(LINUX_2_4) || defined(LINUX_2_6)
       	        DEBUG_EVENT("%s: Initializing for SMP\n",wandev->name);
#endif
#if defined(__LINUX__)
       	       	spin_lock_init(&card->wandev.lock);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
		WAN_SPIN_LOCK_INIT(&card->wandev.lock);
#endif

		/* request an interrupt vector - note that interrupts may be shared */
		/* when using the S514 PCI adapter */
#if defined(__LINUX__)
		if(request_irq(irq, sdla_isr, 
		      (card->type == SDLA_S508) ? 0: SA_SHIRQ, 
		       wandev->name, card)){
			DEBUG_EVENT("%s: Can't reserve IRQ %d!\n", 
					wandev->name, irq);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
 		err = card->hw_iface.set_intrhand(card->hw, sdla_isr, (void*)card);	
		if (err){
			DEBUG_EVENT("%s: Failed set interrupt handler!\n", 
						card->devname);
#endif
			card->hw_iface.down(card->hw);
			err = -EINVAL;
			goto wanpipe_setup_done;
		}
		wandev->irq       = irq;

	}else{
		DEBUG_EVENT("%s: Card Configured %li or Piggybacking %i!\n",
			wandev->name,card->configured,card->wandev.piggyback);
		devconf->comm_port = WANOPT_SEC;
	} 

	if (!card->configured){

		/* Intialize WAN device data space */
		wandev->dma       = 0;
		if(card->type == WANOPT_S50X){ 
			card->hw_iface.getcfg(card->hw, SDLA_IOPORT, &wandev->ioport);
		}else{
			card->hw_iface.getcfg(card->hw, SDLA_CPU, &wandev->S514_cpu_no[0]);
			card->hw_iface.getcfg(card->hw, SDLA_SLOT, &wandev->S514_slot_no);
		}
	}

	switch (card->type){
	case SDLA_S508:
	case SDLA_S514:
		DEBUG_EVENT("%s: Starting Standart Device Init.\n",
				card->devname);
		err = wp_mprot_init(card, devconf);
		break;
	
	case SDLA_ADSL:
		DEBUG_EVENT("%s: Starting ADSL Device Init.\n",
				card->devname);
		/*err = wp_adsl_init(card, NULL);*/
		break;

	case SDLA_AFT:
		DEBUG_EVENT("%s: Starting AFT Hardware Init.\n",
				card->devname);
#if defined(__LINUX__)
		err = wp_xilinx_init(card, devconf);
#else
		err = -EINVAL;
#endif
		break;
	}
	if (err){
		return err;
	}

	/* Register in HDLC device */
	dev = wanpipe_generic_register(card);
	if (dev == NULL){
		return -EINVAL;
	}

#if defined(__LINUX__)
	dev->irq = irq;
#endif

	card->wandev.dev = dev;
	strncpy(ifconf->name, wan_netif_name(dev), WAN_IFNAME_SZ);
	if (card->wandev.new_if){
		err = card->wandev.new_if(&card->wandev, dev, ifconf);
		if (err){
			card->wandev.dev = NULL;
			wanpipe_generic_unregister(dev);
			goto wanpipe_setup_done;
		}
	}

	/* Reserve I/O region and schedule background task */
#if defined(__LINUX__)
       	if(card->type == WANOPT_S50X && !card->wandev.piggyback){
		u16	io_range;
		card->hw_iface.getcfg(card->hw, SDLA_IORANGE, &io_range);
#if defined(LINUX_2_4) || defined(LINUX_2_6)
		if (!request_region(wandev->ioport, io_range, wandev->name)){
			DEBUG_EVENT("%s: Failed to reserve IO region!\n",
						card->devname);
			release_hw(card);
			wandev->state = WAN_UNCONFIGURED;

			card->configured=0;
			sdla_unregister(&card->hw, card->devname);
			if (card->next){
				card->next->next=NULL;
				card->next=NULL;
			}
			err = -EBUSY;
			goto wanpipe_setup_done;
		}
#else
		request_region(wandev->ioport, io_range, wandev->name);
#endif
	}
#endif
	
	card->wandev.critical=0;
	card->wandev.state = WAN_DISCONNECTED;

wanpipe_setup_done:
	wan_free(devconf);
	wan_free(ifconf);
	return err;
	
}

/*================================================================== 
 * wanpipe_find_same_card
 * 
 */ 
static int wanpipe_find_same_card(sdla_t* card)
{
	sdla_t*		nxt_card;
	void*		hwcard;
	void*		nxt_hwcard;
	u16		adapter_type = 0x00;

	if (test_bit(0,&card->configured))
		return 0;
	
	/* Get HW card pointer */
	if (card->hw_iface.get_hwcard == NULL || 
	    card->hw_iface.get_hwcard(card->hw, &hwcard)){
		return 0;
	}
	
	card->hw_iface.getcfg(card->hw, SDLA_ADAPTERTYPE, &adapter_type);
	if (!(adapter_type == S5144_ADPTR_1_CPU_T1E1) ||
	     (adapter_type == S5147_ADPTR_2_CPU_T1E1) ||
	     (adapter_type == S5148_ADPTR_1_CPU_T1E1)){
		return 0;
	}

	/* Check for already loaded card with the same IO port, Bus no and IRQ 
	 * If found, copy its hardware configuration and use its
	 * resources (i.e. piggybacking)
	 */
	
	for (nxt_card = card_list; nxt_card; nxt_card = nxt_card->list){

		if (nxt_card == card)
			continue;
	
		if (nxt_card->wandev.state == WAN_UNCONFIGURED)
			continue;

		if (nxt_card->hw_iface.get_hwcard(nxt_card->hw, &nxt_hwcard)){
			continue;
		}
		if (nxt_hwcard == hwcard){
			nxt_card->same_card = card;
			card->same_card = nxt_card;
			return 0;
		}
	}

	return 0;
}

/*================================================================== 
 * configure_s514_card
 * 
 * For a S514 adapter, check for a possible configuration error in that
 * we are loading an adapter in the same slot as a previously loaded S514
 * card.
 */ 


static int check_conflicts(sdla_t* card)
{
	sdla_t*		nxt_card;

	if (test_bit(0,&card->configured))
		return 0;
	
	/* Check for already loaded card with the same IO port, Bus no and IRQ 
	 * If found, copy its hardware configuration and use its
	 * resources (i.e. piggybacking)
	 */
	
	for (nxt_card=card_list;nxt_card;nxt_card=nxt_card->list) {

		if (nxt_card == card)
			continue;

		if (nxt_card->wandev.state == WAN_UNCONFIGURED)
			continue;

		/* Bug Fix:
		 * If we are using auto PCI slot detection:
		 *   assume that new pci slot is the same as 
		 *   the pci slot of the already configured card
		 *   ie. (nxt_card).
		 * 
		 * The reason for this is, when the pci slot is found
		 * the card->hw.PCI_slot_no is updated with the
		 * found slot number 
		 *
		 * Thus, do not use the PCI_slot_no to detect a
		 * piggyback condition if the auto_pci_cfg option
		 * is enabled.
		 */
		if (nxt_card->hw == card->hw){
		    
			if (nxt_card->next == NULL){
				
				card->wandev.irq = nxt_card->wandev.irq;
	
				/* The master could already be running, we must
				 * set this as a critical area */
				/*lock_adapter_irq(&nxt_card->wandev.lock,&smp_flags);*/
				nxt_card->next = card;
				card->next = nxt_card;

				card->wandev.piggyback = WANOPT_YES;

				/* We must initialise the piggiback spin lock here
				 * since isr will try to lock card->next if it
				 * exists */
#if defined(__LINUX__)
				spin_lock_init(&card->wandev.lock);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
				WAN_SPIN_LOCK_INIT(&card->wandev.lock);
#endif

				/*unlock_adapter_irq(&nxt_card->wandev.lock,&smp_flags);*/

			}
		}
	}

	return 0;
}

/*============================================================================
 * Shut down WAN link driver. 
 * o shut down adapter hardware
 * o release system resources.
 *
 * This function is called by the router when device is being unregistered or
 * when it handles ROUTER_DOWN IOCTL.
 */
static int shutdown (wan_device_t* wandev)
{
	sdla_t*		card = NULL;
	int err=0;

	/* sanity checks */
	if ((wandev == NULL) || (wandev->private == NULL)){
		return -EFAULT;
	}
		
	if (wandev->state == WAN_UNCONFIGURED){
		return 0;
	}
	card = wandev->private;

	wandev->state = WAN_UNCONFIGURED;

	set_bit(PERI_CRIT,(void*)&wandev->critical);
	
	/* In case of piggibacking, make sure that 
         * we never try to shutdown both devices at the same
         * time, because they depend on one another */

	if (card->disable_comm){
		card->disable_comm(card);
	}

	wandev->state = WAN_UNCONFIGURED;

	/* Release Resources */
	release_hw(card);

	wandev->state = WAN_UNCONFIGURED;

        /* only free the allocated I/O range if not an S514 adapter */
	if (card->type == SDLA_S508 && !card->configured){
		u16	io_range;
		card->hw_iface.getcfg(card->hw, SDLA_IORANGE, &io_range);
#if defined(__LINUX__)
              	release_region(wandev->ioport, io_range);
#endif
	}

	if (!card->configured){
		sdla_unregister(&card->hw, card->devname);
		card->hw = NULL;
	      	if (card->next){
			sdla_unregister(&card->next->hw, card->next->devname);
			card->next->hw = NULL;
			card->next->next=NULL;
			card->next=NULL;
		}
	}
	
	clear_bit(PERI_CRIT,(void*)&wandev->critical);
	
	return err;
}

static void release_hw (sdla_t *card)
{
	sdla_t *nxt_card;

	/* Check if next device exists */
	if (card->next){
		nxt_card = card->next;
		/* If next device is down then release resources */
		if (nxt_card->wandev.state == WAN_UNCONFIGURED){
			if (card->wandev.piggyback){
				/* If this device is piggyback then use
                                 * information of the master device 
				 */
				DEBUG_EVENT("%s: Piggyback shutting down\n",card->devname);
				if (card->hw_iface.down){
					card->hw_iface.down(card->next->hw);
				}
				if (card->wandev.config_id != WANCONFIG_BSC && 
				    card->wandev.config_id != WANCONFIG_POS){ 
#if defined(__LINUX__)
       					free_irq(card->wandev.irq, card->next);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
					card->hw_iface.restore_intrhand(card->hw);
#endif
				}
				card->configured = 0;
				card->next->configured = 0;
				card->wandev.piggyback = 0;

			}else{
				/* Master device shutting down */
				DEBUG_EVENT("%s: Master shutting down\n",
						card->devname);
				if (card->hw_iface.down){
					card->hw_iface.down(card->hw);
				}
				if (card->wandev.config_id != WANCONFIG_BSC && 
				    card->wandev.config_id != WANCONFIG_POS){ 
#if defined(__LINUX__)
					free_irq(card->wandev.irq, card);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
					card->hw_iface.restore_intrhand(card->hw);
#endif
				}
				card->configured = 0;
				card->next->configured = 0;
			}

		}else{
			DEBUG_EVENT("%s: Device still running state=%i\n",
				nxt_card->devname,nxt_card->wandev.state);

			card->configured = 1;
		}
	}else{
		DEBUG_EVENT("%s: Master shutting down\n",card->devname);

		if (card->hw_iface.down){
			card->hw_iface.down(card->hw);
		}
		if (card->wandev.config_id != WANCONFIG_BSC && 
		    card->wandev.config_id != WANCONFIG_POS &&
		    card->wandev.config_id != WANCONFIG_DEBUG){
#if defined(__LINUX__)
       			free_irq(card->wandev.irq, card);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
			card->hw_iface.restore_intrhand(card->hw);
#endif
		}
		card->configured = 0;
	}
	return;
}


/****** Driver IOCTL Handlers ***********************************************/

static int 
wanpipe_ioctl(netdevice_t* dev, struct ifreq* ifr, int cmd)
{
	sdla_t*			card;
	wanpipe_common_t*	common;
	struct if_settings*	ifsettings;
	sdla_te_cfg_t		te_cfg;
	wan_dev_conf_t		devconf;
	void*			org_ifr_data = NULL;
	int 			err = 0;

	if (wan_netif_priv(dev) == NULL){
		DEBUG_EVENT("%s: Device private structure is NULL!\n",
				wan_netif_name(dev));
		return -EINVAL;
	}
	common = wan_netif_priv(dev);
	card = common->card;
	if (card == NULL){
		DEBUG_EVENT("%s: Invalid Sangoma device card\n",
				wan_netif_name(dev));
		return -EINVAL;
	}
	switch(cmd){
	case SIOC_WANPIPE_HWPROBE:
		err = wanpipe_ioctl_hwprobe(dev, ifr->ifr_data);
		break;

	case SIOC_WANPIPE_DUMP:
		err = wanpipe_ioctl_dump(card, ifr->ifr_data);
		break;

	case SIOC_WANPIPE_DEVICE:
		ifsettings = (struct if_settings*)ifr->ifr_data;
		switch (ifsettings->type) {
		case IF_IFACE_T1:
		case IF_IFACE_E1:
			if (!IS_TE1(card->wandev.te_cfg)){
				DEBUG_EVENT("%s: Current card is not T1/E1 card!\n",
						card->devname);
				return -EINVAL;
			}
			err = WAN_COPY_FROM_USER(&te_cfg,
						ifsettings->ifs_te1,
						sizeof(te_cfg));
			if (err){
				DEBUG_EVENT("%s: Failed to copy from user space (%d)\n",
						card->devname, __LINE__);
				return -ENOMEM;
			}
			DEBUG_TEST("%s: Set T1/E1 parameters!\n", 
					card->devname);
			DEBUG_TEST("%s: Media %d\n", 
					card->devname, te_cfg.media);
			DEBUG_TEST("%s: LCODE %d\n", 
					card->devname, te_cfg.lcode);
			DEBUG_TEST("%s: FRAME %d\n",
					card->devname, te_cfg.frame);
			DEBUG_TEST("%s: ACTIVE_CH %lX\n",
					card->devname, te_cfg.active_ch);
			DEBUG_TEST("%s: TECLOCK %d\n",
					card->devname, te_cfg.te_clock);
			sdla_te_copycfg(card, &te_cfg);
			/* If Media type changed, update interface value (!) */
			card->wandev.interface = 
				(card->wandev.te_cfg.media == WANOPT_MEDIA_T1)? 
					WANOPT_V35 : WANOPT_RS232;
			/* Only for T1/E1 card!
			 * If media type is change, we need reconfigure
			 * both CPUs. */
			if (card->type != SDLA_AFT && card->same_card){
				sdla_t*	same_card = card->same_card;
				if (card->wandev.te_cfg.media != 
						same_card->wandev.te_cfg.media){
					DEBUG_EVENT("%s: T1/E1 media type is different for %s on the same card!\n",
							card->devname,
							same_card->devname);
					DEBUG_EVENT("%s: Change media type to %s and set default configuration!\n",
							same_card->devname,
							IS_T1_CARD(card) ? "T1" : "E1");
							
					sdla_te_default_cfg(same_card, 
						&same_card->wandev.te_cfg, 
						card->wandev.te_cfg.media);
				}
			}
			break;
		case IF_IFACE_SYNC_SERIAL:
			err = WAN_COPY_FROM_USER(&devconf,
						ifsettings->ifs_sync,
						sizeof(devconf));
			if (err){
				DEBUG_EVENT("%s: Failed to copy from user space (%d)\n",
						card->devname, __LINE__);
				return -ENOMEM;
			}
			if (!IS_TE1(card->wandev.te_cfg)){
				if (devconf.mask & WANHDLC_CONF_INTERFACE){
					card->wandev.interface = 
							devconf.interface;
				}
				if (devconf.mask & WANHDLC_CONF_CLOCKING){
					if (devconf.bps){
						card->wandev.clocking = WANOPT_INTERNAL;
						card->wandev.bps = devconf.bps;
					}else{
						card->wandev.clocking = WANOPT_EXTERNAL;
						card->wandev.bps = 0;
					}
				}
			}
			if (devconf.mask & WANHDLC_CONF_MTU){
				card->wandev.mtu = devconf.mtu;
			}
			if (devconf.mask & WANHDLC_CONF_UDPPORT){
				card->wandev.udp_port = devconf.udp_port;
			}
			if (devconf.mask & WANHDLC_CONF_TTL){
				card->wandev.ttl = devconf.ttl;
			}
			break;
		default:
#if defined(__FreeBSD__) || defined(__OpenBSD__)
			return 1;/*err = sppp_ioctl(dev, SIOCSIFFLAGS, data);*/
#elif defined(__LINUX__)
			org_ifr_data = ifr->ifr_data;
			memcpy(&ifr->ifr_settings,
				ifr->ifr_data,
				sizeof(struct if_settings));
			switch(ifr->ifr_settings.type){
                        case IF_PROTO_PPP:
                        case IF_PROTO_CISCO:
                        case IF_PROTO_FR:
                                common->proto = ifr->ifr_settings.type;
                                break;
                        }
			if (common->proto == IF_PROTO_PPP){
				hdlc_device* hdlc = dev_to_hdlc(dev);
				((wanpipe_common_t*)dev->priv)->prot_ptr = 
						(netdevice_t*)&hdlc->state.ppp.pppdev;
#if defined(LINUX2_6)
				hdlc->state.ppp.syncppp_ptr = (struct ppp_device*)
						((wanpipe_common_t*)dev->priv)->prot_ptr;
#endif
			}
			err = hdlc_ioctl(dev, ifr, SIOCWANDEV);
			memcpy(org_ifr_data,
				&ifr->ifr_settings,
				sizeof(struct if_settings));
			ifr->ifr_data = org_ifr_data;
#endif
			break;
		}
		break;

	default:	
#if defined(__FreeBSD__) || defined(__OpenBSD__)
		return 1;	/*err = sppp_ioctl(dev, cmd, data);*/
#elif defined(__LINUX__)
		err = hdlc_ioctl(dev, ifr, cmd);
#endif
		break;
	}
	return err;
}


static int wanpipe_ioctl_hwprobe(netdevice_t* dev, void* u_def)
{
	sdla_t*		card = NULL;
	wanlite_def_t	def;
	unsigned char*	str;
	int 		err;	

	card = ((wanpipe_common_t*)wan_netif_priv(dev))->card;
	memset(&def, 0, sizeof(wanlite_def_t));
	if (card->hw_iface.get_hwprobe == NULL){
		return -EINVAL;
	}
	err = card->hw_iface.get_hwprobe(
			card->hw, 
			(card->type == SDLA_AFT) ? 
					0 : card->u.c.comm_port,
			(void**)&str);
	if (err){
		return -EINVAL;
	}
	str[strlen(str)] = '\0';
	memcpy(def.hwprobe, str, strlen(str));
	/* Get interface configuration */
	if (IS_TE1(card->wandev.te_cfg)){
		if (IS_T1_CARD(card)){
			def.iface = IF_IFACE_T1;
		}else{
			def.iface = IF_IFACE_E1;
		}
		memcpy(&def.te_cfg, &card->wandev.te_cfg, sizeof(sdla_te_cfg_t));
	}else{
		def.iface = IF_IFACE_SYNC_SERIAL;
		def.devconf.bps = card->wandev.bps;
	}
	def.devconf.mtu = card->wandev.mtu;
	def.devconf.ttl = card->wandev.ttl;
	def.devconf.udp_port = card->wandev.udp_port;

	err = WAN_COPY_TO_USER(u_def, &def, sizeof(def));
	if (err){
		DEBUG_EVENT("%s: Failed to copy to user space (%d)\n",
				card->devname, __LINE__);
		return -ENOMEM;
	}
	return 0;
}

static int wanpipe_ioctl_dump(sdla_t* card, void* u_dump)
{
	sdla_dump_t	dump;
        void*		data;
        u32		memory;
        int		err = 0;

	err = WAN_COPY_FROM_USER(&dump, u_dump, sizeof(sdla_dump_t));
	if (err){
                return -EFAULT;
	}

        card->hw_iface.getcfg(card->hw, SDLA_MEMORY, &memory);
        if ((dump.magic != WANPIPE_MAGIC) ||
            (dump.offset + dump.length > memory)){
                return -EINVAL;
        }

        data = wan_malloc(dump.length);
        if (data == NULL){
                return -ENOMEM;
        }

        card->hw_iface.peek(card->hw, dump.offset, data, dump.length);

	err = WAN_COPY_TO_USER(dump.ptr, data, dump.length);
        if (err){
		DEBUG_EVENT("%s: Failed to copy to user space (%d)\n",
				card->devname, __LINE__);
                err = -EFAULT;
        }
        wan_free(data);
        return err;
}

/******* Miscellaneous ******************************************************/

static int wanpipe_get_default_cfg(sdla_t* card, wandev_conf_t** devconf, wanif_conf_t** ifconf)
{
	wandev_conf_t*	devconf_new;
	wanif_conf_t*	ifconf_new;
	int 		err = 0;
	
	devconf_new = wan_malloc(sizeof(wandev_conf_t));
	if (devconf_new == NULL){
		DEBUG_EVENT("%s: Failed to create default configuration!\n",
				card->devname);
		return -ENOMEM;
	}
	ifconf_new = wan_malloc(sizeof(wanif_conf_t));
	if (ifconf_new == NULL){
		DEBUG_EVENT("%s: Failed to create default configuration!\n",
				card->devname);
		wan_free(devconf_new);
		return -ENOMEM;
	}
	if (!card->hw_iface.is_te1(card->hw)){
		DEBUG_EVENT("%s: Set default T1/E1 config arg!\n",
				card->devname);
		sdla_te_default_cfg(card, &devconf_new->te_cfg, WANOPT_MEDIA_T1);
		devconf_new->interface = 
				(devconf_new->te_cfg.media == WANOPT_MEDIA_T1) ? 
						WANOPT_V35 : WANOPT_RS232;
	}else if (!card->hw_iface.is_56k(card->hw)){
		DEBUG_EVENT("%s: Set default 56K config arg!\n",
				card->devname);
		sdla_56k_default_cfg(card, &devconf_new->te_cfg);
		devconf_new->interface 		= WANOPT_V35;
	}else{
		devconf_new->interface 		= WANOPT_V35;
	}
	devconf_new->clocking		= WANOPT_EXTERNAL;
	devconf_new->comm_port		= WANOPT_PRI;
	devconf_new->ttl		= 255;
	devconf_new->udp_port		= 0;
	devconf_new->ignore_front_end_status	= WANOPT_NO;
	devconf_new->bps		= 0;
	switch (card->type){
	case SDLA_S508:
	case SDLA_S514:
		DEBUG_EVENT("%s: Set default S508/S514 config arg!\n",
				card->devname);
		wp_mprot_default_devcfg(card, devconf_new);
		wp_mprot_default_ifcfg(card, ifconf_new);
		break;
		
	case SDLA_ADSL:
		DEBUG_EVENT("%s: Set default ADSL config arg!\n",
				card->devname);
		break;

	case SDLA_AFT:
		DEBUG_EVENT("%s: Set default AFT config arg!\n",
				card->devname);
#if defined(__LINUX__)
		wp_xilinx_default_devcfg(card, devconf_new);
		wp_xilinx_default_ifcfg(card, ifconf_new);
#else
		err = -EINVAL;
#endif
		break;

	default:
		DEBUG_EVENT("%s: ERROR, invalid card type 0x%0X!\n",
				card->devname, card->type);
		err = -EINVAL;
		break;
	}

	if (err){
		wan_free(devconf_new);
		wan_free(ifconf_new);
		return err;
	}
	*devconf = devconf_new;
	*ifconf = ifconf_new;
	return 0;

}

/*============================================================================
 * SDLA Interrupt Service Routine.
 * o acknowledge SDLA hardware interrupt.
 * o call protocol-specific interrupt service routine, if any.
 */
#if defined(__FreeBSD__) || defined(__OpenBSD__)
STATIC irqreturn_t sdla_isr (void* dev_id)
#elif defined(__LINUX__)
STATIC irqreturn_t sdla_isr (int irq, void* dev_id, struct pt_regs *regs)
#endif
{
#define	card	((sdla_t*)dev_id)
	
	if (card && card->type == SDLA_AFT){

#if defined(__FreeBSD__) || defined(__OpenBSD__)
		wan_spin_lock(&card->wandev.lock);
#elif defined(__LINUX__)
		spin_lock(&card->wandev.lock);
#endif
		if (card->isr){
			card->isr(card);
		}
#if defined(__FreeBSD__) || defined(__OpenBSD__)
		wan_spin_unlock(&card->wandev.lock);
#elif defined(__LINUX__)
		spin_unlock(&card->wandev.lock);
#endif
	
		WAN_IRQ_RETVAL(IRQ_HANDLED);
	}

	if (card && card->type == SDLA_ADSL){
		unsigned long flags;
#if defined(__FreeBSD__) || defined(__OpenBSD__)
		WAN_SPIN_LOCK_IRQSAVE(&card->wandev.lock);
#elif defined(__LINUX__)
		spin_lock_irqsave(&card->wandev.lock,flags);
#endif
		if (card->isr){
			card->isr(card);
		}
#if defined(__FreeBSD__) || defined(__OpenBSD__)
		WAN_SPIN_UNLOCK_IRQSAVE(&card->wandev.lock);
#elif defined(__LINUX__)
		spin_unlock_irqrestore(&card->wandev.lock,flags);
#endif
		WAN_IRQ_RETVAL(IRQ_HANDLED);
	}

	if(card->type == SDLA_S514) {	/* handle interrrupt on S514 */
                u32 int_status;
                u16 CPU_no;
                unsigned char card_found_for_IRQ;

                card->hw_iface.getcfg(card->hw, SDLA_CPU, &CPU_no);
		if (card->hw_iface.read_int_stat){
			card->hw_iface.read_int_stat(card->hw, &int_status);
		}

		/* check if the interrupt is for this device */
		if(!((unsigned char)int_status & (IRQ_CPU_A | IRQ_CPU_B))){
			WAN_IRQ_RETVAL(IRQ_NONE);
		}

		/* if the IRQ is for both CPUs on the same adapter, */
		/* then alter the interrupt status so as to handle */
		/* one CPU at a time */
		if(((unsigned char)int_status & (IRQ_CPU_A | IRQ_CPU_B))
			== (IRQ_CPU_A | IRQ_CPU_B)) {
			int_status &= (CPU_no == SDLA_CPU_A) ?
				~IRQ_CPU_B : ~IRQ_CPU_A;
		}

		card_found_for_IRQ = 0;
		/* check to see that the CPU number for this device */
		/* corresponds to the interrupt status read */
		switch (CPU_no) {
			case SDLA_CPU_A:
				if((unsigned char)int_status & IRQ_CPU_A){
					card_found_for_IRQ = 1;
				}
			break;

			case SDLA_CPU_B:
				if((unsigned char)int_status & IRQ_CPU_B){
					card_found_for_IRQ = 1;
				}
			break;
		}

		/* exit if the interrupt is for another CPU on the */
		/* same IRQ */
		if(!card_found_for_IRQ){
			WAN_IRQ_RETVAL(IRQ_NONE);
		}

		if (!card || 
		   (card->wandev.state == WAN_UNCONFIGURED && !card->configured)){
				DEBUG_EVENT("Received IRQ %d for CPU #%c\n",
					card->wandev.irq, SDLA_GET_CPU(CPU_no));
				DEBUG_EVENT("IRQ for unconfigured adapter\n");
				if (card->hw_iface.intack){
					card->hw_iface.intack(card->hw, int_status);
				}
				WAN_IRQ_RETVAL(IRQ_NONE);
		}

		if (card->in_isr) {
			DEBUG_EVENT("%s: interrupt re-entrancy on IRQ %d\n",
				card->devname, card->wandev.irq);
			if (card->hw_iface.intack){
				card->hw_iface.intack(card->hw, int_status);
			}
			WAN_IRQ_RETVAL(IRQ_NONE);
		}

#if defined(__FreeBSD__) || defined(__OpenBSD__)
		/* FIXME: SMP */
#elif defined(__LINUX__)
		spin_lock(&card->wandev.lock);
#endif
		if (card->next){
#if defined(__FreeBSD__) || defined(__OpenBSD__)
			/* FIXME: SMP */
#elif defined(__LINUX__)
			spin_lock(&card->next->wandev.lock);
#endif
		}
				
		if (card->hw_iface.intack){
			card->hw_iface.intack(card->hw, int_status);
		}

		if (card->isr){
			card->isr(card);
			clear_bit(0,&card->spurious);
		}
		
		if (card->next && card->next->isr){
			card->next->isr(card->next);
			clear_bit(0,&card->next->spurious);
		}

		if (card->next){
#if defined(__FreeBSD__) || defined(__OpenBSD__)
			/* FIXME: SMP */
#elif defined(__LINUX__)
			spin_unlock(&card->next->wandev.lock);
#endif
		}
#if defined(__FreeBSD__) || defined(__OpenBSD__)
		/* FIXME: SMP */
#elif defined(__LINUX__)
		spin_unlock(&card->wandev.lock);
#endif
	
	}else{		/* handle interrupt on S508 adapter */

		if (!card || ((card->wandev.state == WAN_UNCONFIGURED) && !card->configured))
			WAN_IRQ_RETVAL(IRQ_NONE);

		if (card->in_isr) {
			DEBUG_EVENT("%s: interrupt re-entrancy on IRQ %d!\n",
				card->devname, card->wandev.irq);
			WAN_IRQ_RETVAL(IRQ_NONE);
		}

#if defined(__FreeBSD__) || defined(__OpenBSD__)
			/* FIXME: SMP */
#elif defined(__LINUX__)
		spin_lock(&card->wandev.lock);
#endif
		if (card->next){
#if defined(__FreeBSD__) || defined(__OpenBSD__)
			/* FIXME: SMP */
#elif defined(__LINUX__)
			spin_lock(&card->next->wandev.lock);
#endif
		}
	
		if (card->hw_iface.intack){
			card->hw_iface.intack(card->hw, 0x00);
		}
		if (card->isr)
			card->isr(card);
		
		if (card->next){
#if defined(__FreeBSD__) || defined(__OpenBSD__)
			/* FIXME: SMP */
#elif defined(__LINUX__)
			spin_unlock(&card->next->wandev.lock);
#endif
		}
#if defined(__FreeBSD__) || defined(__OpenBSD__)
			/* FIXME: SMP */
#elif defined(__LINUX__)
		spin_unlock(&card->wandev.lock);
#endif

	}

	WAN_IRQ_RETVAL(IRQ_HANDLED);

#undef	card
}

/*============================================================================
 * This routine is called by the protocol-specific modules when network
 * interface is being open.  The only reason we need this, is because we
 * have to call MOD_INC_USE_COUNT, but cannot include 'module.h' where it's
 * defined more than once into the same kernel module.
 */
void wanpipe_open (sdla_t* card)
{
	++card->open_cnt;
#if !defined(LINUX_2_6)
	MOD_INC_USE_COUNT;
#endif	
}

/*============================================================================
 * This routine is called by the protocol-specific modules when network
 * interface is being closed.  The only reason we need this, is because we
 * have to call MOD_DEC_USE_COUNT, but cannot include 'module.h' where it's
 * defined more than once into the same kernel module.
 */
void wanpipe_close (sdla_t* card)
{
	--card->open_cnt;
#if !defined(LINUX_2_6)
	MOD_DEC_USE_COUNT;
#endif
}


void debug_print_udp_pkt(unsigned char *data,int len,char trc_enabled, char direction)
{
	int i,res;
	DEBUG_EVENT("\n");
	DEBUG_EVENT("%s UDP PACKET: ",direction?"RX":"TX");
	for (i=0; i<sizeof(wan_udp_pkt_t); i++){
		if (i==0){
			DEBUG_EVENT("\n");
			DEBUG_EVENT("IP PKT: ");
		}
		if (i==sizeof(iphdr_t)){
			DEBUG_EVENT("\n");
			DEBUG_EVENT("UDP PKT: ");
		}
		if (i==sizeof(iphdr_t)+sizeof(udphdr_t)){
			DEBUG_EVENT("\n");
			DEBUG_EVENT("MGMT PKT: ");
		}
		if (i==sizeof(iphdr_t)+sizeof(udphdr_t)+sizeof(wan_mgmt_t)){
			DEBUG_EVENT("\n");
			DEBUG_EVENT("CMD PKT: ");
		}
		
		if (trc_enabled){
			if (i==sizeof(iphdr_t)+sizeof(udphdr_t)+
			       sizeof(wan_mgmt_t)+sizeof(wan_cmd_t)){
				DEBUG_EVENT("\n");
				DEBUG_EVENT("TRACE PKT: ");
			}
			if (i==sizeof(iphdr_t)+sizeof(udphdr_t)+
			       sizeof(wan_mgmt_t)+sizeof(wan_cmd_t)+
			       sizeof(wan_trace_info_t)){

				DEBUG_EVENT("\n");
				DEBUG_EVENT("DATA PKT: ");
			}

			res=len-(sizeof(iphdr_t)+sizeof(udphdr_t)+
			       sizeof(wan_mgmt_t)+sizeof(wan_cmd_t)+sizeof(wan_trace_info_t));
		
			res=(res>10)?10:res;

			if (i==sizeof(iphdr_t)+sizeof(udphdr_t)+
			       sizeof(wan_mgmt_t)+sizeof(wan_cmd_t)+sizeof(wan_trace_info_t)+res){
				break;
			}
			
		}else{
	
			if (i==sizeof(iphdr_t)+sizeof(udphdr_t)+sizeof(wan_mgmt_t)+sizeof(wan_cmd_t)){
				DEBUG_EVENT("\n");
				DEBUG_EVENT("DATA PKT: ");
			}

			res=len-(sizeof(iphdr_t)+sizeof(udphdr_t)+
			       sizeof(wan_mgmt_t)+sizeof(wan_cmd_t));
		
			res=(res>10)?10:res;

			if (i==sizeof(iphdr_t)+sizeof(udphdr_t)+
			       sizeof(wan_mgmt_t)+sizeof(wan_cmd_t)+res){
				break;
			}
		}

		DEBUG_EVENT("%02X ",*(data+i));
	}
	DEBUG_EVENT("\n");
}

#if defined(__LINUX__)
int wan_reg_api(sdla_t *card, void *chan_ptr, void *dev, void (*func)(unsigned long data))
{
	wanpipe_common_t *chan = (wanpipe_common_t*)chan_ptr;

	if (test_and_set_bit(WAN_API_INIT,&chan->used)){
		DEBUG_EVENT("%s: Error: Failed to initialize API!\n",
				card->devname);
		return -EINVAL;
	}

	DEBUG_EVENT("%s: Initializing API\n",
			card->devname);

	chan->sk=NULL;
	chan->state = WAN_CONNECTING;
	chan->dev = dev;
	chan->api_state=0;
	
	IFQ_INIT(&chan->rx_queue,10);

	WAN_TASKLET_INIT((&chan->task),0,func,(unsigned long)chan);

	set_bit(WAN_API_INIT,&chan->used);
	return 0;
}

int wan_unreg_api(sdla_t *card, void *chan_ptr)
{
	wanpipe_common_t *chan = (wanpipe_common_t*)chan_ptr;
	
	if (!test_and_clear_bit(WAN_API_INIT,&chan->used)){
		DEBUG_EVENT("%s: Error: Failed to unregister API!\n",
				card->devname);
		return -EINVAL;
	}

	DEBUG_EVENT("%s: Unregistering API\n",
			card->devname);

	chan->state = WAN_DISCONNECTED;
	WAN_TASKLET_KILL((&chan->task));
	wan_skb_queue_purge(&chan->rx_queue);	

	return 0;
}

void wan_trigger_api_bh(void *chan_ptr)
{
	wanpipe_common_t *chan = (wanpipe_common_t*)chan_ptr;
	WAN_TASKLET_SCHEDULE((&chan->task));
}

void wan_api_bh_end(void *chan_ptr)
{
	wanpipe_common_t *chan = (wanpipe_common_t*)chan_ptr;
	WAN_TASKLET_END((&chan->task));
}

int wan_api_enqueue_skb(void *chan_ptr,struct sk_buff *skb)
{
	wanpipe_common_t *chan = (wanpipe_common_t*)chan_ptr;

	if (wan_skb_queue_len(&chan->rx_queue) > MAX_API_RX_QUEUE){
		return -EBUSY;
	}
	wan_skb_queue_tail(&chan->rx_queue,skb);

	if (wan_skb_queue_len(&chan->rx_queue) >= MAX_API_RX_QUEUE){
		return 1;
	}

	return 0;
}

struct sk_buff * wan_api_dequeue_skb(void *chan_ptr)
{
	wanpipe_common_t *chan = (wanpipe_common_t*)chan_ptr;
	return wan_skb_dequeue(&chan->rx_queue);
}
#endif
/****** End *********************************************************/
